
import datetime

from .tstamp import serialize_tstamp, deserialize_tstamp_w
from .stdsvcdef import IMsgOStream, IHandshakeService, ISetValueService, IFindTagHandler, IFindTagService, IEchoService, ITagUpdateHandler, ISubscribeService, ISubscribeAllService, ITagIterHandler, ITagIterService
from .stddef import VOLCANO_PROTOCOL_VERSION_STR


def msg_get_w(msg: dict, param_name: str, test_type=None, optional: bool = False, default=None):
    assert isinstance(msg, dict), msg
    assert isinstance(param_name, str), param_name

    if param_name not in msg:
        if optional:
            return default
        raise Warning('Message "{}" does not have obligatory property "{}"'.format(msg, param_name))

    val = msg[param_name]
    if test_type is not None and not isinstance(val, test_type):
        raise Warning('Property "{}={}" should be {} in message {}'.format(param_name, val, test_type, msg))

    return val


def msg_get_tstamp_w(msg: dict, param_name: str, optional: bool = False) -> (None, datetime.datetime):
    s = msg_get_w(msg, param_name, str, optional, None)
    return deserialize_tstamp_w(s)


class HandshakeService(IHandshakeService, IMsgOStream):
    protocol_version = VOLCANO_PROTOCOL_VERSION_STR

    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        # self.transport_: IMsgOStream = transport
        self.transport_ = transport

    def salute(self, inst_name: str) -> None:
        assert isinstance(inst_name, str), inst_name
        self.transport_.push_single_message({'c': "hello", 'name': inst_name, 'protocol': self.protocol_version})

    def push_single_message(self, msg: dict) -> None:
        if msg['c'] != 'welcome':
            raise Exception('Expected welcome from server, got {}'.format(msg))


class SingleThreadEchoService(IEchoService, IMsgOStream):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport
        self.transport_ = transport
        self.echo_cbs_ = []  # [(counter, callback, *args, **kwargs), ...]
        self.echo_cbs_ctr_ = 0

    def echo(self, callback, *args, **kwargs) -> None:
        self.echo_cbs_ctr_ += 1
        self.echo_cbs_.append((self.echo_cbs_ctr_, callback, args, kwargs))

        self.transport_.push_single_message({'c': 'echo', 'ctr': self.echo_cbs_ctr_})

    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg

        if not self.echo_cbs_:
            raise Exception('Echo Request/response mismatch')

        ctr = msg['ctr']

        requested_ctr, cb, args, kwargs = self.echo_cbs_.pop(0)
        if requested_ctr != ctr:
            raise Exception('Echo Request/response mismatch. Requested {}, got {}'.format(requested_ctr, ctr))

        cb(*args, **kwargs)


class SingleThreadFindTagService(IFindTagService, IMsgOStream):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport
        self.transport_ = transport
        self.find_tag_cbs_ = []

    def find_tag(self, tag_name_or_id: (int, str), handler: IFindTagHandler, user_data=None) -> None:
        assert isinstance(tag_name_or_id, (int, str)), tag_name_or_id
        assert isinstance(handler, IFindTagHandler), handler

        self.find_tag_cbs_.append((tag_name_or_id, handler, user_data))
        self.transport_.push_single_message({'c': 'find', 'tag': tag_name_or_id})

    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg

        assert self.find_tag_cbs_

        tag_name_or_id, handler, user_data = self.find_tag_cbs_.pop(0)

        if msg['c'] == 'find_ok':
            tag_id = msg_get_w(msg, 'id', test_type=int)
            tag_name = msg_get_w(msg, 'name', test_type=str)
            vt = msg_get_w(msg, 'type', test_type=str)

            handler.on_find_tag_ok(tag_id, tag_name, vt, user_data)
        else:
            assert msg['c'] == 'find_err'
            handler.on_find_tag_err(tag_name_or_id, user_data)


class SingleThreadTagIterService(ITagIterService, IMsgOStream):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport
        self.transport_ = transport
        # self.handler_: ITagIterHandler = None
        self.handler_ = None

    def first(self, handler: ITagIterHandler):
        assert isinstance(handler, ITagIterHandler), handler

        self.handler_ = handler
        self.transport_.push_single_message({'c': 'first'})

    def next(self, tag_name_or_id: (int, str), handler: ITagIterHandler):
        assert isinstance(tag_name_or_id, (int, str)), tag_name_or_id
        assert isinstance(handler, ITagIterHandler), handler

        self.handler_ = handler
        self.transport_.push_single_message({'c': 'next', 'tag': tag_name_or_id})

    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg

        handler = self.handler_
        self.handler_ = None    # сбросить надо имеено здесь, так как в handler.on_next_tag с большой вероятностью будет вызван .next()
        assert handler

        if msg['c'] == 'next_tag':
            tag_id = msg_get_w(msg, 'id', test_type=int)
            tag_name = msg_get_w(msg, 'name', test_type=str)
            vt = msg_get_w(msg, 'type', test_type=str)
            atts = msg_get_w(msg, 'atts')

            handler.on_next_tag(tag_id, tag_name, vt, atts)

        elif msg['c'] == 'next_end':
            handler.on_next_end()
        else:
            assert msg['c'] == 'next_err'
            handler.on_next_err()


class PipeMarker(IMsgOStream):
    def __init__(self, channel: IMsgOStream, channel_uid):
        self.out_channel_ = channel
        self.uid_ = channel_uid

    def push_single_message(self, msg: dict) -> None:
        msg['ch'] = self.uid_
        self.out_channel_.push_single_message(msg)


# Поскольку данный сервис не имеет собственного кэша сообщений, он является многопоточным (транспорт в свою очередь тоже должен быть МП)
class MultiThreadSetValueService(ISetValueService):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        # self.transport_: IMsgOStream = transport
        self.transport_ = transport

    def set_tag(self, tag_name_or_id: (str, int), value, quality: int = 0, tstamp: ('datetime.datetime', None) = None) -> None:
        assert isinstance(tag_name_or_id, (str, int)), tag_name_or_id
        assert isinstance(quality, int), quality
        assert tstamp is None or isinstance(tstamp, datetime.datetime), tstamp

        req = {'c': 'set', 'tag': tag_name_or_id, 'v': value}
        if quality:
            req['q'] = quality
        if tstamp:
            req['t'] = serialize_tstamp(tstamp)

        self.transport_.push_single_message(req)


class SingleThreadSubscribeService(ISubscribeService):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        # self.transport_: IMsgOStream = transport
        self.transport_ = transport
        self.subs_ = {}  # {name_or_id:(handler, *args, **kwargs), ...}

    def subscribe(self, tag_name_or_id: (int, str), send_tstamp: bool, handler: ITagUpdateHandler) -> None:
        assert isinstance(tag_name_or_id, (int, str)), tag_name_or_id
        assert isinstance(send_tstamp, bool), send_tstamp
        assert isinstance(handler, ITagUpdateHandler), handler

        assert tag_name_or_id not in self.subs_, tag_name_or_id

        self.subs_[tag_name_or_id] = handler

        self.transport_.push_single_message({'c': 'sub', 'tag': tag_name_or_id, 'tstamp': send_tstamp})

    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg
        assert msg['c'] == 'upd', msg

        tag_name_or_id = msg_get_w(msg, 'tag', test_type=(int, str))
        val_raw = msg_get_w(msg, 'v')
        quality = msg_get_w(msg, 'q', test_type=int)
        tstamp_n = msg_get_tstamp_w(msg, 't', optional=True)

        assert tag_name_or_id in self.subs_, tag_name_or_id
        handler = self.subs_[tag_name_or_id]
        handler.on_tag_updated(tag_name_or_id, val_raw, quality, tstamp_n)


class SingleThreadSubscribeAllService(ISubscribeAllService):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        # self.transport_: IMsgOStream = transport
        self.transport_ = transport
        self.handler_ = None

    def subscribe_all(self, send_tstamp: bool, use_tag_id: bool, handler: ITagUpdateHandler) -> None:
        assert isinstance(send_tstamp, bool), send_tstamp
        assert isinstance(use_tag_id, bool), use_tag_id
        assert isinstance(handler, ITagUpdateHandler), handler

        assert self.handler_ is None, 'Duplicated subscribe All call'

        # self.handler_: ITagUpdateHandler = handler
        self.handler_ = handler

        self.transport_.push_single_message({'c': 'sub', 'tag': '*', 'tstamp': send_tstamp, 'ref': 'id' if use_tag_id else 'name'})

    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg
        assert msg['c'] == 'upd', msg

        tag_name_or_id = msg_get_w(msg, 'tag', test_type=(int, str))
        val_raw = msg_get_w(msg, 'v')
        quality = msg_get_w(msg, 'q', test_type=int)
        tstamp_n = msg_get_tstamp_w(msg, 't', optional=True)

        assert self.handler_
        self.handler_.on_tag_updated(tag_name_or_id, val_raw, quality, tstamp_n)


class SingleThreadAllService(IMsgOStream, IHandshakeService, ISetValueService, IEchoService, IFindTagService, ITagIterService,  # pylint: disable=too-many-ancestors
                             ISubscribeService, ISubscribeAllService):

    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        # self.transport_: IMsgOStream = transport
        self.transport_ = transport

        self.services_ = {
            'hs': HandshakeService(PipeMarker(transport, 'hs')),
            'set': MultiThreadSetValueService(transport),
            'echo': SingleThreadEchoService(PipeMarker(transport, 'echo')),
            'find': SingleThreadFindTagService(PipeMarker(transport, 'find')),
            'next': SingleThreadTagIterService(PipeMarker(transport, 'next')),
            'sub': SingleThreadSubscribeService(PipeMarker(transport, 'sub')),
            'suball': SingleThreadSubscribeAllService(PipeMarker(transport, 'suball')),
        }

    def salute(self, inst_name: str) -> None:
        self.services_['hs'].salute(inst_name)

    def set_tag(self, tag_name_or_id: (str, int), value, quality: int = 0, tstamp: ('datetime.datetime', None) = None) -> None:
        self.services_['set'].set_tag(tag_name_or_id, value, quality, tstamp)

    def echo(self, callback, *args, **kwargs) -> None:
        self.services_['echo'].echo(callback, *args, **kwargs)

    def find_tag(self, tag_name_or_id: (str, int), handler: IFindTagHandler, user_data=None) -> None:
        self.services_['find'].find_tag(tag_name_or_id, handler, user_data)

    def first(self, handler: ITagIterHandler):
        self.services_['next'].first(handler)

    def next(self, tag_name_or_id: (int, str), handler: ITagIterHandler):
        self.services_['next'].next(tag_name_or_id, handler)        # pylint: disable=not-callable

    def subscribe(self, tag_name_or_id: (int, str), send_tstamp: bool, handler: ITagUpdateHandler) -> None:
        self.services_['sub'].subscribe(tag_name_or_id, send_tstamp, handler)

    def subscribe_all(self, send_tstamp: bool, use_tag_id: bool, handler: ITagUpdateHandler) -> None:
        self.services_['suball'].subscribe_all(send_tstamp, use_tag_id, handler)

    # as soon as this method is called from transport, it doesnt swallow exceptions
    def push_single_message(self, msg: dict) -> None:
        assert isinstance(msg, dict), msg

        assert 'ch' in msg
        ch_name = msg['ch']

        assert ch_name in self.services_
        self.services_[ch_name].push_single_message(msg)


'''
class SingleThreadOnDemandServices(IMsgOStream):
    def __init__(self, transport: IMsgOStream):
        assert isinstance(transport, IMsgOStream), transport

        #self.transport_: IMsgOStream = transport
        self.transport_ = transport
        self.services_ = {}

    @property
    def svc_set(self) -> ISetValueService:
        if 'set' not in self.services_:
            self.services_['set'] = MultiThreadSetValueService(self.transport_)
        return self.services_['set']

    @property
    def svc_find(self) -> IFindTagService:
        if 'find' not in self.services_:
            self.services_['find'] = SingleThreadFindTagService(PipeMarker(self.transport_, 'find'))
        return self.services_['find']

    @property
    def svc_echo(self) -> IEchoService:
        if 'echo' not in self.services_:
            self.services_['echo'] = SingleThreadEchoService(PipeMarker(self.transport_, 'echo'))
        return self.services_['echo']

    @property
    def svc_next(self) -> ITagIterService:
        if 'next' not in self.services_:
            self.services_['next'] = SingleThreadTagIterService(PipeMarker(self.transport_, 'next'))
        return self.services_['next']

    @property
    def svc_sub(self) -> ISubscribeService:
        if 'sub' not in self.services_:
            self.services_['sub'] = SingleThreadSubscribeService(PipeMarker(self.transport_, 'sub'))
        return self.services_['sub']

    @property
    def svc_sub_all(self) -> ISubscribeAllService:
        if 'suball' not in self.services_:
            self.services_['suball'] = SingleThreadSubscribeAllService(PipeMarker(self.transport_, 'suball'))
        return self.services_['suball']
    """
    def svc(self, svc_name: str):
        assert isinstance(svc_name, str), svc_name
        assert svc_name in self.services_, (svc_name, self.services_)

        return self.services_[svc_name]

    def add_svc(self, svc_name: str, svc_type):
        assert isinstance(svc_name, str), svc_name
        assert svc_name not in self.services_, svc_name

        self.services_[svc_name] = svc_type(PipeMarker(self.transport_, svc_name))
    """
'''     # pylint: disable=pointless-string-statement



import struct

# volcano
from ..lib.stddef import ValueType
from ..lib.bin import bytes_to_str
from ..lib.vector import Vector
from ..lib import stdsvcdef
from ..lib.xml_reader import XmlReader, LoadException, load_xml_file_le
from ..lib.modbus import MB_FORMATS_DICT, MbException

# locals
from .mb_rules import RegRuleNS


MB_EXC_ILLEGAL_DATA_ADDRESS = 0x02
MB_EXC_ILLEGAL_DATA_VALUE = 0x03


class SharedReg(stdsvcdef.IFindTagHandler, stdsvcdef.ITagUpdateHandler):
    def __init__(self, xml_node, device, addr_offset: int):
        p = XmlReader(xml_node)

        self.device_ = device
        self.log = device.log
        self.name_ = p.get_str('name')

        # addr & format
        # (str_id, size_words, min|None, max|None, struct_format)
        self.fmt_tpl_ = p.get_dic('fmt', MB_FORMATS_DICT)
        a, fmt_size_words, *r = self.fmt_tpl_       # pylint: disable=unused-variable

        addr = p.get_int('addr', min_val=0, max_val=(0xffff - fmt_size_words - addr_offset + 1)) + addr_offset

        self.vec_ = Vector(addr, size=fmt_size_words)
        # можно и float, но надо быть осторожным. int(big_int * 1.0) != big_int! потому что при умножении
        # на 1.0 значение преобразуется во float, где на больших порядках (как int64) точность ниже 1.
        # У меня ошибка составляла 11
        self.k_ = p.get_int('k', 1)
        self.value_ = None
        self.exists_ = False

    def exists(self):
        return self.exists_

    def sync(self, find_svc: stdsvcdef.IFindTagService, sub_svc: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(find_svc, stdsvcdef.IFindTagService), find_svc
        assert isinstance(sub_svc, stdsvcdef.ISubscribeService), sub_svc

        tag_full_name = self.device_.branch() + '.' + self.name_
        find_svc.find_tag(tag_full_name, handler=self, user_data=sub_svc)

    def on_find_tag_ok(self, tag_id: int, tag_name: str, vt: str, user_data: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(tag_id, int), tag_id
        assert isinstance(tag_name, str), tag_name
        assert isinstance(vt, str), vt
        assert isinstance(user_data, stdsvcdef.ISubscribeService), user_data

        assert not self.exists_

        if ValueType(vt) in (ValueType.VT_INT, ValueType.VT_FLOAT, ValueType.VT_BOOL):
            self.exists_ = True
            user_data.subscribe(tag_name, send_tstamp=False, handler=self)
        else:
            self.log.warning('Cant share tag %s [%s]. Only support int,float,bool', tag_name, vt)

    def on_find_tag_err(self, tag_name_or_id: (int, str), user_data: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(tag_name_or_id, str), tag_name_or_id  # мы искали по имени
        assert isinstance(user_data, stdsvcdef.ISubscribeService), user_data

        assert not self.exists_

        self.log.warn('Mapped tag not found: %s', tag_name_or_id)

    def on_tag_updated(self, tag_name_or_id: (int, str), val_raw, quality: int, tstamp_n: ('datetime.datetime', None)):  # pylint: disable=unused-argument
        assert isinstance(tag_name_or_id, str), tag_name_or_id  # мы подписывались по имени
        assert isinstance(quality, int), quality

        assert self.exists_

        self.log.debug('Got update %s=%s [%s]', tag_name_or_id, val_raw, quality)

        if quality == 0:
            self.value_ = val_raw
        else:
            self.value_ = None if self.device_.ForceNsIfBadQuality else val_raw

    def map_mbex(self, data_bytes, req_vec):
        if self.exists_ and req_vec.intersects(self.vec_):
            if not req_vec.contains(self.vec_):
                raise MbException(MB_EXC_ILLEGAL_DATA_ADDRESS)

            # val: (int, float)
            if self.value_ is None:
                if self.device_.RuleForNsReg == RegRuleNS.Exception:
                    raise MbException(MB_EXC_ILLEGAL_DATA_VALUE, '')
                else:
                    val = 0
            else:
                if isinstance(self.value_, bool):
                    val = 1 if self.value_ else 0
                else:
                    val = self.value_ if self.k_ == 1 else self.value_ * self.k_

            fmt_id, fmt_size_words, fmt_min_n, fmt_max_n, fmt_struct_format, val_type = self.fmt_tpl_   # pylint: disable=unused-variable
            if fmt_min_n is not None and val < fmt_min_n:
                val = fmt_min_n
            elif fmt_max_n is not None and val > fmt_max_n:
                val = fmt_max_n
            elif not isinstance(val, val_type):
                val = int(val) if val_type == int else float(val)

            struct.pack_into(fmt_struct_format, data_bytes, 2 * (self.vec_.s - req_vec.s), val)


class SharedBit(stdsvcdef.IFindTagHandler, stdsvcdef.ITagUpdateHandler):
    def __init__(self, xml_node, device, addr_offset: int):
        p = XmlReader(xml_node)

        self.device_ = device
        self.log = device.log
        self.name_ = p.get_str('name')

        self.addr_ = p.get_int('addr', min_val=0, max_val=0xffff - addr_offset) + addr_offset
        self.invert_ = p.get_bool('invert', False)
        self.value_ = None
        self.exists_ = False

    def exists(self):
        return self.exists_

    def sync(self, find_svc: stdsvcdef.IFindTagService, sub_svc: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(find_svc, stdsvcdef.IFindTagService), find_svc
        assert isinstance(sub_svc, stdsvcdef.ISubscribeService), sub_svc

        tag_full_name = self.device_.branch() + '.' + self.name_
        find_svc.find_tag(tag_full_name, handler=self, user_data=sub_svc)

    def on_find_tag_ok(self, tag_id: int, tag_name: str, vt: str, user_data: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(tag_id, int), tag_id
        assert isinstance(tag_name, str), tag_name
        assert isinstance(vt, str), vt
        assert isinstance(user_data, stdsvcdef.ISubscribeService), user_data

        assert not self.exists_

        if ValueType(vt) in (ValueType.VT_INT, ValueType.VT_FLOAT, ValueType.VT_BOOL):
            self.exists_ = True
            user_data.subscribe(tag_name, send_tstamp=False, handler=self)
        else:
            self.log.warning('Cant share tag %s [%s]. Only support int,float,bool', tag_name, vt)

    def on_find_tag_err(self, tag_name_or_id: (int, str), user_data: stdsvcdef.ISubscribeService) -> None:
        assert isinstance(tag_name_or_id, str), tag_name_or_id  # мы искали по имени
        assert isinstance(user_data, stdsvcdef.ISubscribeService), user_data

        assert not self.exists_

        self.log.warn('Mapped tag not found: %s', tag_name_or_id)

    def on_tag_updated(self, tag_name_or_id: (int, str), val_raw, quality: int, tstamp_n: ('datetime.datetime', None)):  # pylint: disable=unused-argument
        assert isinstance(tag_name_or_id, str), tag_name_or_id  # мы подписывались по имени
        assert isinstance(quality, int), quality

        assert self.exists_

        self.log.debug('Got update %s=%s [%s]', tag_name_or_id, val_raw, quality)

        if quality == 0:
            self.value_ = val_raw
        else:
            self.value_ = None if self.device_.ForceNsIfBadQuality else val_raw

    def map_mbex(self, bits, req_vec):
        if self.exists_ and req_vec.contains_value(self.addr_):
            if self.value_ is None:
                raise MbException(MB_EXC_ILLEGAL_DATA_VALUE, '')
        
            bits[self.addr_ - req_vec.s] = bool(self.value_) != self.invert_  # this is xor


class Device:
    def __init__(self, xml_node, parent_log):
        p = XmlReader(xml_node)

        self.branch_ = p.get_str('branch')
        self.slave_nb_ = p.get_int('slaveNb', min_val=0, max_val=0xff)
        self.log = parent_log.getChild(str(self.slave_nb_))
        self.addr_offset_ = p.get_int('addrOffset', min_val=0, max_val=10000, default=0)
        self.regs_ = []
        self.bits_ = []

        # что делать, если пришел апдейт, в котором quality!=0, но значение не none:
        #   ForceNsIfBadQuality==true  => в value записывается none
        #   ForceNsIfBadQuality==false => value оставляется как есть
        self.ForceNsIfBadQuality = True                     
        self.RuleForNsReg = RegRuleNS.Exception            # что делать, если регистр NS
        # self.RuleForMinMaxReg = RegRuleMinMax.MinMax       # что делать, если регистр вне зоны minmax модасовского фрейма

        self.RuleForNsReg = p.get_dic('ns', {
            'exception': RegRuleNS.Exception, 
            'zero': RegRuleNS.Zero
            }, default_value=RegRuleNS.Exception)

        template_name = p.get_str('template')
        template_xml = load_xml_file_le(template_name)

        for tag_node in template_xml.getroot():
            if tag_node.tag == 'Tag':
                p2 = XmlReader(tag_node)
                is_coil = p2.get_bool('coil', False)
                if is_coil:
                    self.bits_.append(SharedBit(tag_node, self, self.addr_offset_))
                else:
                    self.regs_.append(SharedReg(tag_node, self, self.addr_offset_))
            else:
                raise LoadException('Unknown node name: {}'.format(tag_node.tag), tag_node)

    def branch(self):
        return self.branch_

    def slave_nb(self):
        return self.slave_nb_

    def sync(self, find_svc: stdsvcdef.IFindTagService, sub_svc: stdsvcdef.ISubscribeService, echo_svc: stdsvcdef.IEchoService) -> None:
        assert isinstance(find_svc, stdsvcdef.IFindTagService), find_svc
        assert isinstance(sub_svc, stdsvcdef.ISubscribeService), sub_svc
        assert isinstance(echo_svc, stdsvcdef.IEchoService), echo_svc

        for x in self.regs_:
            x.sync(find_svc, sub_svc)
        for x in self.bits_:
            x.sync(find_svc, sub_svc)

        echo_svc.echo(self.on_sync_complete)

    def on_sync_complete(self):
        self.regs_ = tuple(filter(lambda x: x.exists(), self.regs_))
        self.bits_ = tuple(filter(lambda x: x.exists(), self.bits_))

    # can raise MbException
    def mb_read_words_mbex(self, addr, nb_words, resp_bytes):
        req_vec = Vector(addr, size=nb_words)
        for reg in self.regs_:
            reg.map_mbex(resp_bytes, req_vec)

        self.log.debug('Read words %s/%s => data[%s]', addr, nb_words, bytes_to_str(resp_bytes))

    def mb_read_bits_mbex(self, addr, nb_bits, resp_bits):
        req_vec = Vector(addr, size=nb_bits)
        for bit in self.bits_:
            bit.map_mbex(resp_bits, req_vec)

        self.log.debug('Read bits %s/%s => data[%s]', addr, nb_bits, resp_bits)



from twisted.internet import reactor, protocol


class VolcanoTwistedFactory(protocol.ClientFactory):
    nb_attempts = 3
    nb_attempts_made = 0
    reconnect_pause = 3.0
    log = None
    client = None

    def clientConnectionFailed(self, connector, reason):
        self.nb_attempts_made += 1
        if self.nb_attempts_made < self.nb_attempts:
            if self.log:
                self.log.warn(
                    'Connection to Volcano {} failed (attempt {}/{}): {}. Try again in {} secs...'.format(
                        connector.getDestination(),
                        self.nb_attempts_made,
                        self.nb_attempts,
                        reason.getErrorMessage(),
                        self.reconnect_pause))
            reactor.callLater(self.reconnect_pause, connector.connect)  # pylint: disable=no-member
        else:
            if self.log:
                self.log.error('Connection to Volcano {} failed (attempt {}/{}): {}. Abort!'.format(
                    connector.getDestination(),
                    self.nb_attempts_made,
                    self.nb_attempts,
                    reason.getErrorMessage()))
            reactor.stop()      # pylint: disable=no-member

    def clientConnectionLost(self, connector, reason):
        if self.log:
            self.log.error('Connection to volcano-core lost: {}'.format(reason.getErrorMessage()))
        if reactor.running:     # pylint: disable=no-member
            reactor.stop()      # pylint: disable=no-member

    def buildProtocol(self, addr):
        if self.client:
            return self.client
        else:
            return super().buildProtocol(addr)



#!/usr/bin/python3
import socket
import select

from ..lib.xml_reader import XmlReader

from . import channel_base


class TcpChannel(channel_base.BaseChannel):
    def __init__(self, node):
        p = XmlReader(node)

        self.ip_addr_ = p.get_str('tcp.ipAddr')
        self.tcp_port_ = p.get_int('tcp.tcpPort')

        self.sock_ = None

    def open(self, timeout_sec: float = 5.0, log=None) -> bool:
        self.safe_close(log)

        if log:
            log.debug('Connecting to {}:{} ...'.format(self.ip_addr_, self.tcp_port_))

        try:
            # self.sock_ = socket.socket()
            # self.sock_.connect ((self.ip_addr_, self.tcp_port_))
            self.sock_ = socket.create_connection((self.ip_addr_, self.tcp_port_), timeout=timeout_sec)
        except Exception as e:
            if log:
                log.warning('Could not connect to {}:{} - {}'.format(self.ip_addr_, self.tcp_port_, e))
            self.safe_close(log)
            return False

        if log:
            log.debug('Connected to {}:{}'.format(self.ip_addr_, self.tcp_port_))

        return True

    def safe_close(self, log=None) -> None:
        if self.sock_ is None:
            return
            
        if log:
            log.debug('Shutdown socket...')

        try:
            self.sock_.shutdown(1)
        except Exception as ex:
            if log:
                log.warning('Error shutdown socket: {}'.format(ex))
            
        if log:
            log.debug('Close socket...')
        try:
            self.sock_.close()
        except Exception as ex:
            if log:
                log.warning('Error closing socket: {}'.format(ex))
                
        self.sock_ = None

    def is_open(self) -> bool:
        return self.sock_ is not None

    def write_all_x(self, data: (bytes, bytearray), log=None) -> None:
        assert isinstance(data, (bytes, bytearray)), data

        if not self.is_open():
            raise ConnectionError('Cant write to closed channel')

        try:
            self.sock_.sendall(data)
        except Exception as ex:
            if log:
                log.warn('Error writing socket: %s', ex)
            self.safe_close(log)
            raise ConnectionError('Could not send data: {}'.format(ex))

    def read_x(self, timeout_sec: float, log=None) -> bytes:

        if not self.is_open():
            raise ConnectionError('Cant read from closed channel')

        fd_read, fd_write, fd_err = select.select([self.sock_], [], [], timeout_sec)    # pylint: disable=unused-variable
        if self.sock_ in fd_read:
            try:
                data = self.sock_.recv(1024)
            except Exception as ex: # Выбрасывает исключения. Я вообще не понимаю, откуда инфа, что при ошибках возвр None
                if log:
                    log.warn('Error reading socket: %s', ex)
                self.safe_close(log)
                raise ConnectionError(ex)
                
            if data:
                return data
            else:
                self.safe_close(log)
                raise ConnectionError('Channel closed while reading')
        else:
            return b''



#!/usr/bin/python3

from ..lib.bin import bytes_to_str, check_crc16, get_crc16_as_bytes

from . import base_ctx
from .merk_nzif_common import NackException, BadDataException, RequestResponseMismatch


class MerkNzifCtx(base_ctx.BaseCtx):
    def __init__(self, channel, is_merk: bool):
        super().__init__(channel)

        self.is_merk_ = is_merk

        self.req__reply_sz_no_add_no_crc = None
        self.req__expecting_status_reply = None
        self.req__addr = None
        self.req__comment = None

        self.reply_data = None  # no slave nb, no crc

    def get_nack_comment_n(self, nack_code: int) -> (None, str):
        raise NotImplementedError()

    def data_cb(self, data: (bytes, bytearray)):
        assert isinstance(data, (bytes, bytearray)), data

        reply_sz = self.req__reply_sz_no_add_no_crc + 1 + 2
        if len(data) > reply_sz:
            raise BadDataException(
                "{}: Reply is too long (expected {} bytes with crc, got {}): {}".format(self.req__comment, reply_sz,
                                                                                        len(data),
                                                                                        bytes_to_str(data)))

        if len(data) < reply_sz:
            # nack?
            if not self.req__expecting_status_reply and len(data) == 4 and data[0] == self.req__addr and data[1] != 0 and check_crc16(data):
                code = data[1]
                code_comment = self.get_nack_comment_n(code)
                if code_comment:
                    raise NackException('{}: Nack (code={} - {})'.format(self.req__comment, code, code_comment), code)
                else:
                    raise NackException('{}: Nack (code={})'.format(self.req__comment, code), code)
            else:
                return None  # need more data

        if not check_crc16(data):
            raise BadDataException('{}: crc error in {}'.format(self.req__comment, bytes_to_str(data)))

        if data[0] != self.req__addr:
            raise RequestResponseMismatch(
                '{}: Slave number mismatch. Request={}; reply={}'.format(self.req__comment, self.req__addr, data[0]))

        if self.req__expecting_status_reply:
            self.reply_data = None
            code = data[1]
            if code != 0:  # nack
                code_comment = self.get_nack_comment_n(code)
                if code_comment:
                    raise NackException('{}: Nack (code={} - {})'.format(self.req__comment, code, code_comment), code)
                else:
                    raise NackException('{}: Nack (code={})'.format(self.req__comment, code), code)
        else:
            self.reply_data = data[1:]

        return True  # return anything but None

    # reply_data_sz: without address and crc
    # always prints exception text
    # closes channel on commfail and baddata
    def send_rcv(self, data_no_crc: (bytes, bytearray), reply_data_sz_n: (None, int), comment: str, log) -> None:
        assert isinstance(data_no_crc, (bytes, bytearray)), data_no_crc
        assert reply_data_sz_n is None or isinstance(reply_data_sz_n, int), reply_data_sz_n
        assert isinstance(comment, str), comment

        # should not append_crc to data because data can immutable bytes
        data = data_no_crc + get_crc16_as_bytes(data_no_crc)

        self.req__comment = comment
        self.req__addr = data[0]
        if reply_data_sz_n is None:  # we expect status reply
            self.req__reply_sz_no_add_no_crc = 1
            self.req__expecting_status_reply = True
        else:
            self.req__reply_sz_no_add_no_crc = reply_data_sz_n
            self.req__expecting_status_reply = False

        self.send_rcv_raw(data, comment, log)
        
    # data has no slave nb and no crc. data(0) gives first byte after slave nb
    def data(self, byte_idx: int = None):
        return self.reply_data if byte_idx is None else self.reply_data[byte_idx]



#!/usr/bin/python3

# Line numbering
# https://stackoverflow.com/questions/6949395/is-there-a-way-to-get-a-line-number-from-an-elementtree-element

# Force python XML parser not faster C accelerators
# because we can't hook the C implementationimport sys
import sys
import os.path

sys.modules['_elementtree'] = None
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-position;     # noqa: E402


class LineNumberingParser(ET.XMLParser):
    def __init__(self, file_name):
        super().__init__()
        self.file_name_ = os.path.basename(os.path.normpath(file_name))

    def _start(self, tag, attr_list):
        element = super()._start(tag, attr_list)

        element.line_number = self.parser.CurrentLineNumber
        element.file_name = self.file_name_

        return element


def get_node_location(node):
    if hasattr(node, 'line_number'):
        assert hasattr(node, 'file_name')
        return '{}|L{}|<{}>'.format(node.file_name, node.line_number, node.tag)
    else:
        return node.tag


def load_xml_file_le(file_name: str):
    try:
        rs = ET.parse(file_name, parser=LineNumberingParser(file_name))
        return rs
    except Exception as e:
        raise LoadException(e)


class LoadException(Exception):
    def __init__(self, msg_or_exc, node=None):
        if node is not None:  # using 'if node' is not recommended - the behaviour will change
            super().__init__('{}: {}'.format(get_node_location(node), msg_or_exc))
        else:
            super().__init__(msg_or_exc)


class XmlReader:
    def __init__(self, xml_node):
        self.node_ = xml_node

    def location(self):
        return self.node_.tag

    def assert_node_name(self, names: (str, tuple, list)) -> None:
        assert isinstance(names, (str, tuple, list)), names

        if self.node_.tag not in ((names,) if isinstance(names, str) else names):   # pylint: disable=superfluous-parens
            raise LoadException('Unknown node name {}. Available: {}'.format(self.node_.tag, names), self.node_)

    def get_str(self, name, default=None, allowed=None) -> str:
        v = self.node_.attrib.get(name, None)
        if v is None:
            if default is None:
                raise LoadException('Attribute "{}" is missing'.format(name), self.node_)
            return default

        if allowed and v not in allowed:
            raise LoadException('Attribute "{}={}" is invalid. Allowed values: {}'.format(name, v, allowed), self.node_)

        return v

    def get_bool(self, name, default=None):
        s = self.node_.attrib.get(name, None)
        if s is None:
            if default is None:
                raise LoadException('Attribute "{}" is missing'.format(name), self.node_)
            return default

        if s == '0':
            return False
        elif s == '1':
            return True
        else:
            raise LoadException('Attribute "{}={}" is not a valid bool. Use "0" and "1"'.format(name, s), self.node_)

    def get_int(self, name, default=None, min_val=None, max_val=None):
        v = self.node_.attrib.get(name, None)
        if v is None:
            if default is None:
                raise LoadException('Attribute "{}" is missing'.format(name), self.node_)
            return default
        try:
            iv = int(v)

            if min_val is not None and iv < min_val:
                raise LoadException('Attribute "{}={}" is less than {}'.format(name, v, min_val), self.node_)

            if max_val is not None and iv > max_val:
                raise LoadException('Attribute "{}={}" is greater than {}'.format(name, v, max_val), self.node_)

            return iv
        except Exception as e:
            raise LoadException('Attribute "{}={}" is not a valid int: {}'.format(name, v, e), self.node_)

    def get_dic(self, name, dic, default_value=None):
        v = self.node_.attrib.get(name, None)
        if v is None:
            if default_value is None:
                raise LoadException('Attribute "{}" is missing'.format(name), self.node_)
            return default_value

        rs = dic.get(v, None)
        if rs is None:
            raise LoadException('Attribute "{}={}" is invalid. Available values: {}'.format(name, v, dic.keys()), self.node_)

        return rs

    def get_time_secs(self, name, default=None, min_val=None, max_val=None):
        s = self.node_.attrib.get(name, None)
        if s is None:
            if default is None:
                raise LoadException('Attribute "{}" is missing'.format(name), self.node_)
            return default

        val_secs = None
        try:
            if s.endswith('ms'):
                val_secs = int(s[:-2]) / 1000.0
            elif s.endswith('s'):
                val_secs = float(s[:-1])
            else:
                raise LoadException(
                    'Attribute "{}={}" is invalid. Value should be in a form "123.45s" or "12345ms" where s=seconds, ms=milliseconds.'.format(
                        name, s), self.node_)
        except ValueError as e:
            raise LoadException('Attribute "{}={}" is not a valid time: {}'.format(name, s, e), self.node_)

        if min_val is not None and val_secs < min_val:
            raise LoadException('Attribute "{}={}" is less than {}'.format(name, val_secs, min_val), self.node_)

        if max_val is not None and val_secs > max_val:
            raise LoadException('Attribute "{}={}" is greater than {}'.format(name, val_secs, max_val), self.node_)

        return val_secs



from setuptools import setup

with open('requirements.txt') as f:
    requires = list(filter(lambda x: not not x, map(lambda x: x.strip(), f.readlines())))

with open('version.txt') as f:
    ver = f.read().strip()

setup(
    name='volcano-x',
    version=ver,
    description='VolcanoX on Python',
    author='Vinogradov D',
    author_email='dgrapes@gmail.com',
    license='MIT',
    packages=['volcano.core', 'volcano.poller', 'volcano.web', 'volcano.mbsrv', 'volcano.lib', 'volcano.twistedclient'],
    package_data={'': ['demo.xml']},
    zip_safe=False,
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



#!/usr/bin/python3
# encoding=utf-8


g_merk_nack_codes = {
    1: 'Недопустимая команда или параметр',
    2: 'Внутренняя ошибка счетчика',
    3: 'Не достаточен уровень доступа для удовлетворения запроса',
    4: 'Внутренние часы счетчика уже корректировались в течение текущих суток',
    5: 'Не открыт канал связи',
}

g_merk_type_defs = {
    'merk230': {},
}



from ..lib import tstamp


from .errors import ServerError, ServerErrorCode


class MsgParser:
    def __init__(self, msg: dict):
        assert isinstance(msg, dict), msg
        self.msg_ = msg

    def get_raw(self, name: str, allow_none: bool = False, optional: bool = False, default=None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional

        if name not in self.msg_:
            if optional:
                return default
            raise ServerError('Parameter {} is missing'.format(name), ServerErrorCode.ProtocolError)

        v = self.msg_[name]
        if v is None:
            if allow_none:
                return None
            raise ServerError('Parameter {} is None'.format(name), ServerErrorCode.ProtocolError)

        return v

    # options should consist of str. None should not be put there
    def get_str(self, name: str, allow_none: bool = False, optional: bool = False, default=None, options: (list, tuple) = None) -> (str, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional
        assert options is None or isinstance(options, (list, tuple)), options

        if name not in self.msg_:
            if optional:
                return default
            raise ServerError('Parameter {} is missing'.format(name), ServerErrorCode.ProtocolError)

        v = self.msg_[name]
        if v is None:
            if allow_none:
                return None
            raise ServerError('Parameter {} is None'.format(name), ServerErrorCode.ProtocolError)

        if not isinstance(v, str):
            raise ServerError('Parameter {}={} should be str, not {}'.format(name, v, type(v)), ServerErrorCode.ProtocolError)

        if options and v not in options:
            raise ServerError('Parameter {}={} is invalid, available values: {}'.format(name, v, options), ServerErrorCode.ProtocolError)

        return v

    def get_bool(self, name: str, allow_none: bool = False, optional: bool = False, default=None) -> (bool, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional

        if name not in self.msg_:
            if optional:
                return default
            raise ServerError('Parameter {} is missing'.format(name), ServerErrorCode.ProtocolError)

        v = self.msg_[name]
        if v is None:
            if allow_none:
                return None
            raise ServerError('Parameter {} is None'.format(name), ServerErrorCode.ProtocolError)

        if not isinstance(v, bool):
            raise ServerError('Parameter {}={} should be bool, not {}'.format(name, v, type(v)), ServerErrorCode.ProtocolError)

        return v

    def get_int(self, name: str, allow_none: bool = False, optional: bool = False, default=None) -> (str, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional

        if name not in self.msg_:
            if optional:
                return default
            raise ServerError('Parameter {} is missing'.format(name), ServerErrorCode.ProtocolError)

        v = self.msg_[name]
        if v is None:
            if allow_none:
                return None
            raise ServerError('Parameter {} is None'.format(name), ServerErrorCode.ProtocolError)

        if not isinstance(v, int):
            raise ServerError('Parameter {}={} should be int, not {}'.format(name, v, type(v)), ServerErrorCode.ProtocolError)

        return v

    def get_datetime(self, name: str, optional: bool = False) -> ('datetime.datetime', None):
        s = self.get_str(name, allow_none=True, optional=optional, default=None)
        try:
            return tstamp.deserialize_tstamp_w(s) if s else None
        except Warning as ex:
            raise ServerError('Parameter {}={} is invalid: {}'.format(name, s, ex), ServerErrorCode.ProtocolError)

    def get_tag(self, name: str, allow_none: bool = False, optional: bool = False, default=None) -> (int, str, None):
        assert isinstance(name, str), name
        assert isinstance(allow_none, bool), allow_none
        assert isinstance(optional, bool), optional

        if name not in self.msg_:
            if optional:
                return default
            raise ServerError('Parameter {} is missing'.format(name), ServerErrorCode.ProtocolError)

        v = self.msg_[name]
        if v is None:
            if allow_none:
                return None
            raise ServerError('Parameter {} is None'.format(name), ServerErrorCode.ProtocolError)

        if not isinstance(v, (int, str)):
            raise ServerError('Parameter {}={} should be (int,str), not {}'.format(name, v, type(v)), ServerErrorCode.ProtocolError)

        return v



#!/usr/bin/python3


class Vector:
    def __init__(self, start: int, end: int = None, size: int = None):
        assert isinstance(start, int), start
        assert end is not None or size is not None
        assert end is None or size is None

        self.s = start
        if end is not None:
            self.e = end
        else:
            self.e = start + size - 1

    def intersects(self, v) -> bool:
        return not (v.s > self.e or self.s > v.e)

    def __eq__(self, v) -> bool:
        return self.s == v.s and self.e == v.e

    def __ne__(self, v) -> bool:
        return self.s != v.s or self.e != v.e

    def contains(self, v) -> bool:
        return self.s <= v.s and self.e >= v.e

    def contains_value(self, v: int) -> bool:
        return self.s <= v <= self.e

    def size(self) -> int:
        return self.e - self.s + 1

    def intersection(self, v):
        s = self.s if self.s > v.s else v.s
        e = self.e if self.e < v.e else v.e

        return Vector(s, end=e)



import subprocess
import shlex
import os.path

from ..lib import variant
from ..lib.stddef import Quality, ValueType
from ..lib import tstamp
from ..lib.xml_reader import LoadException, load_xml_file_le, XmlReader, get_node_location

from .expr import Expr


class Tag:  # pylint: disable=too-many-instance-attributes
    def __init__(self, parent_n: ('Tag', None), short_name: str, tag_id: int, var: variant.Variant = None, q: Quality = Quality.QUALITY_NOT_INIT):
        assert parent_n is None or isinstance(parent_n, Tag), parent_n
        assert isinstance(short_name, str), short_name
        assert isinstance(tag_id, int), tag_id
        assert var is None or isinstance(var, variant.Variant), var
        assert isinstance(q, Quality), q

        self.parent_ = parent_n  # None
        self.id = tag_id  # pylint: disable=invalid-name
        self.short_name_ = short_name
        self.full_name_ = (parent_n.full_name() + '.' + short_name) if parent_n else short_name
        self.children_ = []
        self.attributes = {}

        self.expr = None
        self.expr_deps = []  # for which tags this one is a part of expression

        self.var_ = var or variant.VOID_VALUE
        if not self.var_.is_void():
            self.quality = q
            self.tstamp_utc = tstamp.get_current_tstamp_utc()

    def vqt_dict_nonvoid(self, include_tstamp: bool = True):
        assert not self.is_void(), self
        assert isinstance(include_tstamp, bool), include_tstamp

        if include_tstamp:
            return {'v': self.var_.get_value(), 'q': self.quality, 't': tstamp.serialize_tstamp(self.tstamp_utc)}
        return {'v': self.var_.get_value(), 'q': self.quality}

    def var(self):
        return self.var_

    def is_void(self):
        return self.var_.is_void()

    def has_children(self):
        return len(self.children_) > 0

    def parent(self):
        return self.parent_

    def short_name(self):
        return self.short_name_

    def enum_child_tags(self):
        return self.children_

    def full_name(self):
        return self.full_name_

    def find_child(self, short_name):
        for x in self.children_:
            if x.short_name() == short_name:
                return x
        return None

    def set_strict(self, value, quality: Quality, ts) -> None:
        self.var_.set_value_w(value)
        self.quality = quality
        self.tstamp_utc = ts or tstamp.get_current_tstamp_utc()


class Db:
    def __init__(self, log):
        self.root_tags_ = []
        self.all_tags_map_ = {}
        self.all_tags_list_ = []
        self.log = log

    def nb_tags(self) -> int:
        return len(self.all_tags_list_)

    def find_tag_by_name(self, full_name: str) -> (Tag, None):
        assert isinstance(full_name, str), str

        return self.all_tags_map_.get(full_name, None)

    def find_tag_by_id(self, tag_id: int) -> (Tag, None):
        assert isinstance(tag_id, int), tag_id

        if 0 < tag_id <= len(self.all_tags_list_):
            return self.all_tags_list_[tag_id - 1]
        return None

    def find_tag_by_name_or_id(self, tag_name_or_id: (int, str)) -> (Tag, None):
        assert isinstance(tag_name_or_id, (int, str)), tag_name_or_id

        if isinstance(tag_name_or_id, int):
            return self.find_tag_by_id(tag_name_or_id)
        return self.find_tag_by_name(tag_name_or_id)

    def enum_root_tags(self):
        return self.root_tags_

    def all_tags_list(self):
        return self.all_tags_list_

    def create_tag(self, parent_tag_n: (Tag, None), short_name: str, var: variant.Variant = None, q: Quality = Quality.QUALITY_NOT_INIT) -> Tag:
        assert parent_tag_n is None or isinstance(parent_tag_n, Tag), parent_tag_n
        assert isinstance(short_name, str), short_name
        assert var is None or isinstance(var, variant.Variant), var
        assert isinstance(q, Quality), q

        tag_id = len(self.all_tags_list_) + 1
        tag = Tag(parent_tag_n, short_name, tag_id, var, q)
        if parent_tag_n:
            parent_tag_n.children_.append(tag)
        else:
            self.root_tags_.append(tag)

        self.all_tags_list_.append(tag)
        self.all_tags_map_[tag.full_name()] = tag

        return tag

    def create_tag_by_long_name(self, parent_tag_n: (Tag, None), long_name: str, var: variant.Variant = None, q: Quality = Quality.QUALITY_NOT_INIT) -> Tag:
        assert isinstance(long_name, str), long_name
        assert isinstance(q, Quality), q

        if '.' in long_name:  # short name
            names = long_name.split('.')
            parent_name = '.'.join(names[:-1])
            self_name = names[-1]
            my_parent_tag = self.ensure_branch(parent_tag_n, parent_name)
            return self.create_tag(my_parent_tag, self_name, var, q)
        else:
            return self.create_tag(parent_tag_n, long_name, var, q)

    def ensure_branch(self, root_node_n: Tag, long_name: str) -> Tag:
        assert root_node_n is None or isinstance(root_node_n, Tag), root_node_n
        assert isinstance(long_name, str), long_name

        full_name = (root_node_n.full_name() + '.' + long_name) if root_node_n else long_name
        cr = None
        for n in full_name.split('.'):
            child = cr.find_child(n) if cr else self.find_tag_by_name(n)
            if not child:
                child = self.create_tag(cr, n)
            cr = child
        return cr

    def upload_file(self, file_name: str, parent_tag_n: Tag = None) -> None:
        assert parent_tag_n is None or isinstance(parent_tag_n, Tag), parent_tag_n

        file = load_xml_file_le(file_name)
        for node in file.getroot():
            self.upload_xml_node(node, parent_tag_n)

    def setup_demo_mode(self) -> None:
        demo_db_file_name = os.path.dirname(__file__) + '/demo.xml'
        self.upload_file(demo_db_file_name)

    def upload_xml_node(self, node, parent_tag_n: Tag = None) -> None:  # pylint: disable=too-many-statements
        assert parent_tag_n is None or isinstance(parent_tag_n, Tag), parent_tag_n

        p = XmlReader(node)

        if node.tag == 'Tag':
            vt_str = p.get_str('type', ValueType.VT_VOID)
            try:
                vt = ValueType.parse(vt_str)
            except ValueError:
                raise LoadException('Unknown value type: "{}". Available: {}'.format(vt_str, ValueType.VT_ALL_LIST), node)

            var = variant.Variant.from_vt(vt)
            name = p.get_str('name')
            tag = self.create_tag_by_long_name(parent_tag_n, name, var)

            expr_str = p.get_str('expr', '')
            if expr_str:
                try:
                    expr_inst = Expr.load_w(expr_str, tag, self)
                except Warning as e:
                    raise LoadException('{}'.format(e), node)

                for src in expr_inst.sources():
                    assert isinstance(src, Tag)
                    if src == tag:
                        raise LoadException('Cant use self in expression', node)
                    if src.expr:
                        raise LoadException('Cant use chained expressions', node)
                    src.expr_deps.append(tag)
                tag.expr = expr_inst

            for sub_node in node:
                self.upload_xml_node(sub_node, tag)

        elif node.tag == 'File':
            path = p.get_str('path')
            branch = p.get_str('branch', '')
            target_node = self.ensure_branch(parent_tag_n, branch) if branch else parent_tag_n
            try:
                self.upload_file(path, target_node)
            except LoadException as e:  # file is changed and if exc is raised from inside, we wont get info on original location
                raise LoadException('{}: {}'.format(get_node_location(node), e))

        elif node.tag == 'Proc':
            cmd_line = p.get_str('command')
            self.log.info('Upload tags from process %s ...' % cmd_line)
            args = shlex.split(cmd_line)
            p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)  # stderr > stdout
            ctr = 0
            while True:
                line = p.stdout.readline()
                if not line:
                    break
                s = line.strip().decode('utf-8')
                cmd, name, vt_str = s.split()
                if cmd == 'tag':
                    ctr += 1
                    try:
                        vt = ValueType.parse(vt_str)
                    except ValueError:
                        raise LoadException('Unknown value type: "{}". Available: {}'.format(vt_str, ValueType.VT_ALL_LIST))
                    var = variant.Variant.from_vt(vt)
                    self.create_tag_by_long_name(parent_tag_n, name, var)
                    self.log.debug('Added tag %s [%s]' % (name, vt_str))
                else:
                    self.log.warn('Ignored unknown line from process: %s' % s)

            self.log.info('Upload %d tags from process' % ctr)

        elif node.tag == 'Att':
            if not parent_tag_n:
                raise LoadException('Cant use Attributes on top level', node)

            att_name = p.get_str('n')
            att_value = p.get_str('v', '')

            parent_tag_n.attributes[att_name] = att_value

        else:
            raise LoadException('Unknown node name: "{}"'.format(node.tag), node)

    def resolve_path_w(self, branch_tag_n: Tag, path: str) -> Tag:
        assert branch_tag_n is None or isinstance(branch_tag_n, Tag), branch_tag_n
        assert isinstance(path, str), path

        res = branch_tag_n
        for n in path.split('.'):
            if n == '$up':
                if res:
                    res = res.parent()
                else:
                    raise Warning('Cant resolve path {} from {}: went above root'.format(
                        path, branch_tag_n.full_name() if branch_tag_n else '<root>'))
            else:
                if res:
                    res = res.find_child(n)
                else:
                    res = self.find_tag_by_name(n)

                if not res:
                    raise Warning(
                        'Cant resolve path {} from {}: tag {} does not exist'.format(path,
                                                                                     branch_tag_n.full_name() if branch_tag_n else '<root>',
                                                                                     n))

        if not res:
            raise Warning(
                'Cant resolve path {} from {}: stopped on root'.format(path, branch_tag_n.full_name() if branch_tag_n else '<root>'))

        return res



#!/usr/bin/python3
import sys
import argparse
import logging

from twisted.internet import reactor
from twisted.internet.error import CannotListenError

# volcano
from ..lib.log import configure_arg_parser_for_log, configure_logger
from ..lib.stddef import VOLCANO_DEFAULT_TCP_PORT
from ..lib.xml_reader import load_xml_file_le, XmlReader, LoadException
from ..lib.stdsvc import SingleThreadAllService

from ..twistedclient.twisted_factory import VolcanoTwistedFactory
from ..twistedclient.twisted_client_st import VolcanoTwistedClientST

from . import device
from .mb_server import MbFactory


def configure_my_args(parser):
    parser.add_argument('--core_host', help='Volcano core host', default='localhost')
    parser.add_argument('--core_port', help='Volcano core port', default=VOLCANO_DEFAULT_TCP_PORT, type=int)
    parser.add_argument('-n', '--name', help='Instance name', default='mbsrv')
    parser.add_argument('-f', '--file', help='Config file', default='mbsrv.xml')
    parser.add_argument('-i', '--iface', help='Listen interface', default='')
    parser.add_argument('-p', '--port', help='ModbusTCP port', type=int)
    parser.add_argument('--max_con', help='Max number of connections', type=int, default=2)


def override_xml_env(name: str, xml_val, env_val, default_val, log):
    if not xml_val and not env_val:
        log.info('%s set to default value %s', name, default_val)
        return default_val
    elif xml_val and env_val:
        log.info('%s set to %s because command line has priority over xml (in xml value is %s)', name, env_val, xml_val)
        return env_val
    elif xml_val:
        assert not env_val, env_val
        return xml_val
    else:
        return env_val

# ! LoadException
def load(env, log) -> list:
    file = load_xml_file_le(env.file)

    p = XmlReader(file.getroot())
    devices = []

    env.iface = override_xml_env('Interface', p.get_str('iface', ''), env.iface, '0.0.0.0', log)
    env.port = override_xml_env('Port', p.get_int('port', 0, min_val=1, max_val=0xffff), env.port, 502, log)

    for node in file.getroot():
        if node.tag == 'Device':
            dev = device.Device(node, log)
            devices.append(dev)
        else:
            raise LoadException('Unknown node name: "{}"'.format(node.tag))

    if not devices:
        raise LoadException('No devices configured')

    return devices


class MyLavaClient(VolcanoTwistedClientST):
    def __init__(self, env, devices):
        super().__init__(logging.getLogger(env.name))
        self.env = env
        self.devices = devices
        self.all_svc_ = None

    def connectionMade(self):
        super().connectionMade()

        self.all_svc_ = SingleThreadAllService(self)
        self.all_svc_.salute(self.env.name)

        for dev in self.devices:
            dev.sync(self.all_svc_, self.all_svc_, self.all_svc_)
        self.all_svc_.echo(self.sync_complete)

    def sync_complete(self):
        self.log.info('Start listen: iface=%s port=%s', self.env.iface, self.env.port)
        try:
            reactor.listenTCP(self.env.port, MbFactory(self.devices, self.env), interface=self.env.iface)   # pylint: disable=no-member
        except CannotListenError as ex:
            self.log.error(ex)
            reactor.stop()      # pylint: disable=no-member

    def on_msg_rcvd_no_exc(self, msg: dict) -> None:   # Exceptions not expected
        self.all_svc_.push_single_message(msg)


def main():
    try:
        # Args
        arg_parser = argparse.ArgumentParser()
        configure_my_args(arg_parser)
        configure_arg_parser_for_log(arg_parser)  # let logger add his arguments for later configure_logger() call
        env = arg_parser.parse_args()

        # Logging
        configure_logger(env)
        log = logging.getLogger(env.name)

        try:
            devices = load(env, log)
        except LoadException as ex:
            log.error(ex)
            return 1

        volcano_factory = VolcanoTwistedFactory()
        volcano_factory.log = log
        volcano_factory.client = MyLavaClient(env, devices)

        log.info('Connect to core at %s:%s', env.core_host, env.core_port)
        reactor.connectTCP(env.core_host, env.core_port, volcano_factory)           # pylint: disable=no-member
        reactor.run()                                                               # pylint: disable=no-member

        return 0
    finally:
        logging.shutdown()


if __name__ == '__main__':
    sys.exit(main())



from ..lib.stddef import Quality
from ..lib import stdsvcdef
from ..lib import variant

from . import build_cfg


class ConnectedValue(stdsvcdef.IFindTagHandler):
    def __init__(self, full_name: str, var: variant.Variant, is_obligatory: bool):
        assert full_name and isinstance(full_name, str), full_name
        assert isinstance(var, variant.Variant), var
        assert isinstance(is_obligatory, bool), is_obligatory

        self.full_name_ = full_name
        self.is_obligatory_ = is_obligatory
        self.var_ = var
        self.quality_ = Quality.QUALITY_NOT_INIT

        self.exists_ = False
        self.read_ = False
        self.pipe_set_ = None

        build_cfg.build_tag(None, full_name, var.type_str())

    def is_connected(self):
        return self.exists_

    def mark_unread(self):
        self.read_ = False

    def is_read(self):
        return self.read_

    def sync(self, services: dict):
        assert isinstance(services, dict), services

        assert 'find' in services, services
        assert isinstance(services['find'], stdsvcdef.IFindTagService), services
        services['find'].find_tag(self.full_name_, self)

        assert 'set' in services, services
        assert isinstance(services['set'], stdsvcdef.ISetValueService), services
        self.pipe_set_ = services['set']

    def on_find_tag_ok(self, tag_id: int, tag_name: str, vt: str, user_data):
        if self.var_.type_str() != vt:
            raise Warning('Tag "{}" has mismatched type. Expected {}, in fact {}'.format(self.full_name_, self.var_.type_str(), vt))

        self.exists_ = True

        # init tag with NS before we read it
        self.pipe_set_.set_tag(tag_name, self.var_.get_value(), self.quality_)

    def on_find_tag_err(self, tag_name_or_id, user_data):
        if self.is_obligatory_:
            raise Warning('Tag "{}" not found'.format(self.full_name_))

    def set(self, value, quality: int = 0):
        assert isinstance(quality, int), quality

        self.read_ = True

        if self.is_connected():
            if self.quality_ != quality or self.var_.get_value() != value:  # in Python None==None
                self.var_.set_value_w(value)
                self.quality_ = quality
                self.pipe_set_.set_tag(self.full_name_, value, quality)

    def invalidate(self, quality: int):
        self.set(None, quality)



#!/usr/bin/python3

import argparse
import logging
import json
import os.path
import datetime

from twisted.web.resource import Resource, NoResource
from twisted.web.server import Site
from twisted.web.static import File
from twisted.web.util import redirectTo
from twisted.internet import reactor
from twisted.internet.error import CannotListenError

from ..lib.log import configure_arg_parser_for_log, configure_logger
from ..lib import stdsvcdef
from ..lib.xml_reader import LoadException
from ..lib.stddef import VOLCANO_DEFAULT_TCP_PORT, ValueType
from ..lib import variant
from ..lib.stdsvc import SingleThreadAllService

from ..twistedclient.twisted_factory import VolcanoTwistedFactory
from ..twistedclient.twisted_client_st import VolcanoTwistedClientST


def configure_my_args(parser):
    parser.add_argument('--core_host', help='Volcano core host', default='localhost')
    parser.add_argument('--core_port', help='Volcano core port', default=VOLCANO_DEFAULT_TCP_PORT, type=int)
    parser.add_argument('-n', '--name', help='Instance name', default='web')
    parser.add_argument('-i', '--iface', help='Listen interface', default='')  # twisted.listenTcp (..iface='') defaults to all
    parser.add_argument('-p', '--port', help='Listen port', default=8080, type=int)
    parser.add_argument('-w', '--www', help='Static files', default=None)
    parser.add_argument('-c', '--ctrl', help='Enable remote control', action='store_true')
    #parser.add_argument('-g', '--get', help='Enable remote control over GET verb', action='store_true')


class Tag:
    def __init__(self, parent_none, short_name, var, atts):
        assert isinstance(var, variant.Variant), var

        self.parent_ = parent_none  # can be None
        self.short_name_ = short_name  # always non-empty
        self.var = var
        self.atts = atts
        self.quality = None
        self.tstamp = None
        self.children = []

    def parent(self):
        return self.parent_

    def short_name(self):
        return self.short_name_

    def full_name(self):
        return self.short_name_ if self.parent_ is None else (self.parent_.full_name() + '.' + self.short_name_)


class Db:
    def __init__(self):
        self.all_tags_by_name = {}
        self.root_tags_list = []

    def find(self, full_name):
        return self.all_tags_by_name.get(full_name, None)

    def add(self, tag_name, var, atts):
        tag = None

        names = tag_name.split('.')
        if len(names) == 1:
            tag = Tag(None, tag_name, var, atts)
            self.root_tags_list.append(tag)
        else:
            parent_name = '.'.join(names[0:-1])
            short_name = names[-1]
            parent = self.find(parent_name)
            assert parent
            tag = Tag(parent, short_name, var, atts)
            parent.children.append(tag)

        self.all_tags_by_name[tag.full_name()] = tag

        return tag


class WebException(Exception):
    def __init__(self, message, http_code=400):
        super().__init__(message)
        self.http_code = http_code


class Args:     # pylint: disable=too-few-public-methods
    def __init__(self, req):
        self.req_ = req

    def get_str(self, name, default=None):
        val = self.req_.args.get(name.encode('utf-8'), None)
        if val is None or not val:  # val is none or list[]
            if default is None:
                raise WebException('Missing "{}" argument'.format(name), 400)
            return default

        assert isinstance(val, list)
        return val[0].decode('utf-8')


class VolcanoJsonService(Resource):
    isLeaf = True

    def get_json(self, request):
        pass

    def render_GET(self, request):
        try:
            res = self.get_json(request)    # pylint: disable=assignment-from-no-return
        except WebException as ex:  # other exceptions produce html page with call stack
            request.setHeader(b'content-type', b'text/plain')
            request.setResponseCode(ex.http_code)
            return str(ex).encode('utf-8')

        request.setHeader(b'content-type', b'application/json')
        return json.dumps(res).encode('utf-8')


class CmdInfo(VolcanoJsonService):
    def get_json(self, request):
        return {
            "version": {
                "major": 1,
                "minor": 1,
                "patch": 1,
            },
            "web": {
                "ctrl": g_env.ctrl,
                "remoteStop": False,
            }
        }


class CmdLs(VolcanoJsonService):
    def get_json(self, request):
        args = Args(request)
        tag_name = args.get_str('tag', '')  # throws WebException
        tags_list = None
        if not tag_name:  # list root
            tags_list = g_db.root_tags_list
        else:
            tag = g_db.find(tag_name)
            if not tag:
                raise WebException('Tag "{}" not found'.format(tag_name), 400)
            tags_list = tag.children
        return list(map(lambda x: x.short_name(), tags_list))


class CmdStat(VolcanoJsonService):
    def get_json(self, request):
        args = Args(request)

        tag_name = args.get_str('tag')  # throws WebException
        data = args.get_str('d', '')

        tag = g_db.find(tag_name)
        if not tag:
            raise WebException('Tag "{}" not found'.format(tag_name), 400)

        res = {}

        # title, eu, id, children, parents
        if 't' in data:
            res['type'] = tag.var.type_str()
        if 'T' in data:
            res['title'] = tag.atts.get('title', '')
        if 'n' in data:
            res['name'] = tag.short_name()
        if 'N' in data:
            res['fullName'] = tag.full_name()
        if 'c' in data:
            res['hasChildren'] = len(tag.children) > 0  # pylint: disable=len-as-condition

        return res


class CmdRead(VolcanoJsonService):
    def get_json(self, request):
        args = Args(request)

        tag_name = args.get_str('tag')  # throws WebException
        tag = g_db.find(tag_name)

        if not tag:
            raise WebException('Tag "{}" not found'.format(tag_name), 400)

        if tag.var.is_void():
            raise WebException('Tag "{}" is VOID'.format(tag_name), 400)

        return {
            'v': tag.var.get_value(),
            'q': tag.quality,
            #'t': '{}Z'.format(tag.tstamp.isoformat(timespec='milliseconds')) if tag.tstamp else None,
            # isoformat with timespec is Python 3.6
            't': (tag.tstamp.isoformat() + 'Z') if tag.tstamp else None,
        }
        
        
class CmdWrite(VolcanoJsonService):
    isLeaf = True
    
    def render_GET(self, request):
        '''
            На данный момент управление возможно через GET по следующим причинам:
                - удобно тестировать через браузер
                - сама по себе фича все равно отладочная
        '''
        #request.setHeader(b'content-type', b'text/plain')
        #request.setResponseCode(400)
        #return 'Use POST to set '.encode('utf-8')
        return self.render_POST(request)

    def render_POST(self, request):  # pylint: disable=no-self-use
        try:
            args = Args(request)    # !WebException
            
            tag_name = args.get_str('tag')  # !WebException
            tag_val = args.get_str('val')  # !WebException
            tag = g_db.find(tag_name)

            if not tag:
                raise WebException('Tag "{}" not found'.format(tag_name), 404)

                
            # env.ctrl is not checked because otherwise CmdWrite handler wont be added

            # Convert value
            proper_value = None
            if tag.var.vt() == ValueType.VT_BOOL:
                if tag_val == '1':
                    proper_value = True
                elif tag_val == '0':
                    proper_value = False
                else:
                    raise WebException('Cant convert value "{}" to type {} of tag "{}". Use "1" and "0"'.format(tag_val, tag.var.vt(), tag_name), 400)
            elif tag.var.vt() == ValueType.VT_INT:
                try:
                    proper_value = int(tag_val)
                except ValueError:
                    raise WebException('Cant convert value "{}" to type {} of tag "{}"'.format(tag_val, tag.var.vt(), tag_name), 400)
            elif tag.var.vt() == ValueType.VT_FLOAT:
                try:
                    proper_value = float(tag_val)
                except ValueError:
                    raise WebException('Cant convert value "{}" to type {} of tag "{}"'.format(tag_val, tag.var.vt(), tag_name), 400)
            elif tag.var.vt() == ValueType.VT_STR:
                proper_value = tag_val
            else:
                raise WebException('Tag "{}" is VOID'.format(tag_name), 400)
            
            MyLavaClient.Inst.set_tag(tag_name, proper_value)
            
            request.setHeader(b'content-type', b'text/plain')
            return 'OK'.encode('utf-8')
            
        except WebException as ex:  # other exceptions produce html page with call stack
            request.setHeader(b'content-type', b'text/plain')
            request.setResponseCode(ex.http_code)
            return str(ex).encode('utf-8')


class MyLavaClient(VolcanoTwistedClientST, stdsvcdef.ITagIterHandler, stdsvcdef.ITagUpdateHandler):
    Inst = None

    def __init__(self):
        super().__init__(logging.getLogger(g_env.name))
        self.all_svc_ = None

    def on_msg_rcvd_no_exc(self, msg: dict) -> None:   # Exceptions not expected
        self.all_svc_.push_single_message(msg)

    def connectionMade(self):
        
        MyLavaClient.Inst = self
        
        super().connectionMade()

        self.all_svc_ = SingleThreadAllService(self)
        self.all_svc_.salute(g_env.name)

        self.all_svc_.first(self)

    def on_next_tag(self, tag_id: int, tag_name: str, vt: str, atts: dict):
        g_db.add(tag_name, variant.Variant.from_vt(ValueType(vt)), atts)

        self.all_svc_.next(tag_id, self)

    def on_next_end(self):
        self.log.info('Sync finished')
        self.all_svc_.subscribe_all(send_tstamp=True, use_tag_id=False, handler=self)

        if g_env.www:
            self.log.info('Start listen: iface={} port={} static={}'.format(g_env.iface, g_env.port, g_env.www))
        else:
            self.log.info('Start listen: iface={} port={}'.format(g_env.iface, g_env.port))

        try:
            reactor.listenTCP(g_env.port, g_http_factory, interface=g_env.iface)    # pylint: disable=no-member
        except CannotListenError as ex:
            self.log.error(ex)
            reactor.stop()  # pylint: disable=no-member

    def on_next_err(self):
        raise Warning('Failed enumerating')

    def on_tag_updated(self, tag_name_or_id: (int, str), val_raw, quality: int, tstamp_n: (datetime.datetime, None)):
        tag = g_db.find(tag_name_or_id)
        if tag:
            tag.var.set_value_w(val_raw)
            tag.quality = quality
            tag.tstamp = tstamp_n

    def set_tag(self, tag_name_or_id, tag_val):
        self.all_svc_.set_tag(tag_name_or_id, tag_val)


class MyLavaFactory(VolcanoTwistedFactory):
    protocol = MyLavaClient


class IndexResource(Resource):
    def render_GET(self, request):  # pylint: disable=no-self-use
        return redirectTo(b'/www/index.html', request)


class FaviconResource(Resource):
    def render_GET(self, request):  # pylint: disable=no-self-use
        return redirectTo(b'/www/favicon.ico', request)


class MyWebRoot(Resource):
    index = IndexResource()
    favicon = FaviconResource()

    def getChild(self, path, request):
        if g_env.www and path in (b'', b'index.html'):
            return self.index

        if g_env.www and (path == b'favicon.ico'):
            return self.favicon

        resp = Resource.getChild(self, path, request)
        if isinstance(resp, NoResource):
            g_log.warning('Not found %s', path)
        return resp


if __name__ == '__main__':
    # Args
    arg_parser = argparse.ArgumentParser()
    configure_my_args(arg_parser)
    configure_arg_parser_for_log(arg_parser)  # let logger add his arguments for later configure_logger() call
    g_env = arg_parser.parse_args()

    # Logging
    configure_logger(g_env)
    g_log = logging.getLogger(g_env.name)

    if g_env.www:
        if not os.path.isdir(g_env.www):
            g_log.error('Static files directory (-w -www) not found: %s', g_env.www)
    else:
        my_path = os.path.dirname(__file__)
        assert my_path
        g_env.www = my_path + '/www'

        if os.path.isdir(g_env.www):
            g_log.info('Default www path is used: %s', g_env.www)
        else:
            g_log.error('Default www path not found: %s', g_env.www)

    g_log.info('Remote control is %s', 'ON' if g_env.ctrl else 'OFF')
    

    try:
        g_db = Db()
        g_http_factory = None
        volcano_factory = MyLavaFactory()
        volcano_factory.log = g_log

        root = MyWebRoot()

        if g_env.www:
            root.putChild(b'www', File(g_env.www))

        root.putChild(b'info.json', CmdInfo())
        root.putChild(b'ls.json', CmdLs())
        root.putChild(b'stat.json', CmdStat())
        root.putChild(b'read.json', CmdRead())
        
        if g_env.ctrl:
            root.putChild(b'set', CmdWrite())

        g_http_factory = Site(root)

        g_log.info('Connect to core at %s:%s', g_env.core_host, g_env.core_port)
        reactor.connectTCP(g_env.core_host, g_env.core_port, volcano_factory)   # pylint: disable=no-member
        reactor.run()       # pylint: disable=no-member

    except (LoadException, Warning) as ex:
        g_log.error(ex)

    finally:
        logging.shutdown()



#!/usr/bin/python3
import serial

from ..lib.xml_reader import XmlReader, LoadException

from . import channel_base

g_parity_map = {
    'N': serial.PARITY_NONE,
    'E': serial.PARITY_EVEN,
    'O': serial.PARITY_ODD,
}
g_bytesize_map = {
    '8': serial.EIGHTBITS,
}
g_stopbits_map = {
    '1': serial.STOPBITS_ONE,
}


# throws ValueError
# s = '9600 8n1'
def parse_serial_options(s: str) -> dict:
    assert isinstance(s, str), s

    words = s.split(' ')
    if len(words) != 2:
        raise ValueError('Should be 2 words')

    if len(words[1]) != 3:
        raise ValueError('Second item should be in form 8n1')

    baudrate = int(words[0])    # ! ValueError
    if baudrate not in [9600, 19200, 38400]:
        raise ValueError('Invalid baudrate value: {}'.format(baudrate))

    bytesize = words[1][0]
    if bytesize not in g_bytesize_map:
        raise ValueError('Invalid data bits value: {}'.format(bytesize))

    parity = words[1][1].upper()
    if parity not in g_parity_map:
        raise ValueError('Invalid parity: {}'.format(parity))

    stopbits = words[1][2]
    if stopbits not in g_stopbits_map:
        raise ValueError('Invalid stop bits value: {}'.format(stopbits))

    return {
        'baudrate': baudrate,
        'bytesize': g_bytesize_map[bytesize],
        'parity': g_parity_map[parity],
        'stopbits': g_stopbits_map[stopbits]
    }


class SerialChannel(channel_base.BaseChannel):
    def __init__(self, node):
        p = XmlReader(node)

        self.device_name_ = p.get_str('serial.device')

        params_s = p.get_str('serial.params')
        try:
            self.params_ = parse_serial_options(params_s)
        except ValueError as ex:
            raise LoadException('Parameters are invalid: {}'.format(ex), node)

        self.port_ = serial.Serial()

        self.port_.port = self.device_name_
        self.port_.baudrate = self.params_['baudrate']
        self.port_.parity = self.params_['parity']
        self.port_.bytesize = self.params_['bytesize']
        self.port_.stopbits = self.params_['stopbits']

    def open(self, timeout_sec: float = 5.0, log=None) -> bool:
        self.safe_close(log)

        try:
            self.port_.open()
        except Exception as ex:
            self.safe_close(log)
            log.error('Could not open port %s:%s - %s', self.device_name_, self.params_, ex)
            return False

        return True

    def safe_close(self, log=None) -> None:
        if self.is_open():
            self.port_.close()

    def is_open(self) -> bool:
        # https://stackoverflow.com/questions/41987168/serial-object-has-no-attribute-is-open
        # PySerial had isOpen() function until v3.0
        if float(serial.VERSION) < 3.0:
            return self.port_.isOpen()
        else:
            return self.port_.is_open

    def write_all_x(self, data: (bytes, bytearray), log=None) -> None:
        assert isinstance(data, (bytes, bytearray)), data

        if not self.is_open():
            raise ConnectionError('Cant write to closed channel')

        try:
            self.port_.write(data)
        except Exception as ex:
            self.safe_close(log)
            raise ConnectionError('Could not send data: {}'.format(ex))

    def read_x(self, timeout_sec: float, log=None) -> bytes:
        if not self.is_open():
            raise ConnectionError('Cant read from closed channel')

        self.port_.timeout = timeout_sec
        d = self.port_.read(1)
        return d



#!/usr/bin/python3

from . import merk_nzif_ctx
from . import merk_defs


class MerkCtx(merk_nzif_ctx.MerkNzifCtx):
    def __init__(self, channel):
        super().__init__(channel, True)

    def get_nack_comment_n(self, nack_code):
        return merk_defs.g_merk_nack_codes.get(nack_code, None)



import json 

from ..lib import variant
from ..lib.stddef import Quality
from ..lib.xml_reader import XmlReader, LoadException

from . import time_tools
from . import base_dev
from . import connected_values
from .merk_nzif_common import ProtocolException

MEASURES_ENERGY = (
    ('Eap', 'Ean', 'Erp', 'Ern', 0x00),
)


MERK230_TYPE_DEF = {
    'msr-pqs': [
        {'tags': ['P', 'P1', 'P2', 'P3'], 'f': 0.01, 'bwri': [0, 0, 0], 'pq_sign': 'P'},
        {'tags': ['Q', 'Q1', 'Q2', 'Q3'], 'f': 0.01, 'bwri': [0, 1, 0], 'pq_sign': 'Q'},
        {'tags': ['S', 'S1', 'S2', 'S3'], 'f': 0.01, 'bwri': [0, 2, 0], 'pq_sign': None}
    ],
    'msr-uicos': [
        {'tags': ['Ua', 'Ub', 'Uc'], 'f': 0.01, 'bwri': [1, None, None]},
        {'tags': ['Ia', 'Ib', 'Ic'], 'f': 0.001, 'bwri': [2, None, None]},
        {'tags': ['CosAB', 'CosAC', 'CosBC'], 'f': 0.01, 'bwri': [5, None, None]}
    ],
    'fn_08_11':[
    ]
}

'''
    Пример раздела fn_08_11:
    'fn_08_11':[
        {'tag': 'Ia', 'f': 0.001, 'bwri': [2, 0, 1]},
        {'tag': 'Ib', 'f': 0.001, 'bwri': [2, 0, 2]},
        {'tag': 'Ic', 'f': 0.001, 'bwri': [2, 0, 3]},
    ]
'''

FREQ_TAG_NAME = 'F'
STATUS_TAG_NAME = 'status'


class MerkDevice(base_dev.BaseDevice):

    def __init__(self, my_xml_node, proposed_aux_name, parent_log):
        super().__init__(my_xml_node, proposed_aux_name, parent_log)

        log = self.log()
        
        p = XmlReader(my_xml_node)

        self.addr_ = p.get_int('slaveNb', min_val=0, max_val=0xff)

        # device type
        # self.type_def_ = p.get_dic('type', merk_defs.g_merk_type_defs)
        # Правильно было бы сделать default='', но в этом случае теряется совместимость со старыми проектами, где не было json файла,
        # но в xml был прописан тип 'merk230'
        type_name = p.get_str('type', default='merk230')
        if type_name == 'merk230':
            log.warn('Using builtin Merkury template. To use custom, put json file in work dir and specify the name in "type" parameter. The name should NOT be "merk230" - it is reserved for backward compatability')
            type_def = MERK230_TYPE_DEF
        else:
            try:
                with open(type_name + '.json') as f:
                    type_def = json.loads(f.read())
            except Exception as ex:
                raise LoadException(ex)

        # password
        pwd_str = p.get_str('pwd', '000000')
        if len(pwd_str) == 6:
            self.password_ = bytes([ord(x) for x in pwd_str])
        else:
            pwd_tokens = pwd_str.split(',')
            if len(pwd_tokens) != 6:
                raise Exception('{}: password "{}" is invalid: should be 6 ascii characters (111111) or 6 comma-separated codes (0x1,0x1,0x1,0x1,0x1,0x1)'.format(p.location(), pwd_str))
            
            try:
                self.password_ = bytes([int(tok, 16 if tok.startswith('0x') else 10) for tok in pwd_tokens])
            except:
                raise Exception('{}: password "{}" is invalid: should be 6 ascii characters (111111) or 6 comma-separated codes (0x1,0x1,0x1,0x1,0x1,0x1)'.format(p.location(), pwd_str))

        self.log().debug('Password parsed as {}'.format(self.password_))

        # timeouts
        self.timeout_general_ = time_tools.Timeout(p.get_time_secs('readPeriod', 5.0, min_val=0.0, max_val=600.0), expired=True)

        # Create values
        self.val_online_ = connected_values.ConnectedValue(self.branch('online'), variant.BoolValue(), True)
        self.values_ = {}
        for n1, n2, n3, n4, *r in MEASURES_ENERGY:      # pylint: disable=unused-variable
            for name in (n1, n2, n3, n4):
                self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)

        # self.msr_pqs_ = MERK230_TYPE_DEF['msr-pqs']
        self.msr_pqs_ = type_def.get('msr-pqs', [])
        for x in self.msr_pqs_:
            nX, n1, n2, n3 = x['tags']
            for name in (nX, n1, n2, n3):
                if name:  # nX can be None
                    self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)

        # self.msr_uicos_ = MERK230_TYPE_DEF['msr-uicos']
        self.msr_uicos_ = type_def.get('msr-uicos', [])
        for x in self.msr_uicos_:
            n1, n2, n3 = x['tags']
            for name in (n1, n2, n3):
                self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
        
        # self.fn_08_11_ = MERK230_TYPE_DEF.get('fn_08_11', [])
        self.fn_08_11_ = type_def.get('fn_08_11', [])
        for x in self.fn_08_11_:
            name = x['tag']
            self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
            
        
        self.values_[FREQ_TAG_NAME] = connected_values.ConnectedValue(self.branch(FREQ_TAG_NAME), variant.FloatValue(), is_obligatory=False)
        self.values_[STATUS_TAG_NAME] = connected_values.ConnectedValue(self.branch(STATUS_TAG_NAME), variant.IntValue(), is_obligatory=False)
                

    def sync(self, services: dict):
        assert isinstance(services, dict), services
        super().sync(services)
        self.val_online_.sync(services)
        for k, v in self.values_.items():       # pylint: disable=unused-variable
            v.sync(services)

    def on_channel_closed(self):
        self.log().debug('Channel closed and device going offline')
        self.set_offline()

    def set_offline(self):
        self.val_online_.set(False)
        # self.scaling_ = None
        for k, v in self.values_.items():       # pylint: disable=unused-variable
            v.invalidate(Quality.QUALITY_COMM)

    def work(self, ctx):
        log = self.log()

        if not self.timeout_general_.is_expired():
            return

        try:
            self.open_channel(ctx)
        except (ConnectionError, TimeoutError, ProtocolException) as e:
            log.warning(e)
            self.set_offline()
            self.timeout_general_.start()
            return

        self.val_online_.set(True)
        for name, con_val in self.values_.items():  # pylint: disable=unused-variable
            con_val.mark_unread()

        try:
            comment = 'Read status'
            try:
                self.read_status(ctx, comment)
            except (ConnectionError, TimeoutError, ProtocolException) as ex:
                log.warning('Error reading {}: {}'.format(comment, ex))
                if not ctx.is_open():
                    raise
        
            for n1, n2, n3, n4, arr in MEASURES_ENERGY:
                try:
                    self.read_energy(ctx, arr, 'Read energy {} {} {} {}'.format(n1, n2, n3, n4), (n1, n2, n3, n4))
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    if ctx.is_open():
                        log.warning(e)
                    else:
                        raise

            def calc_bwri(arr) -> int:
                p1, p2, p3 = arr
                res = p1 << 4
                if p2 is not None:
                    res |= p2 << 2
                if p3 is not None:
                    res |= p3
                return res
            

            for x in self.msr_pqs_:
                nX, n1, n2, n3 = x['tags']
                factor = x['f']
                bwri = calc_bwri(x['bwri'])
                pq_sign = x['pq_sign']

                comment = 'Read {} {} {}'.format(n1, n2, n3)
                try:
                    self.read_pqs(ctx, bwri, factor, comment, (nX, n1, n2, n3), pq_sign)
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    log.warning('Error reading {}: {}'.format(comment, e))
                    if not ctx.is_open():
                        raise

            for x in self.msr_uicos_:
                n1, n2, n3 = x['tags']
                factor = x['f']
                bwri = calc_bwri(x['bwri'])

                comment = 'Read {} {} {}'.format(n1, n2, n3)
                try:
                    self.read_uicos(ctx, bwri, factor, comment, (n1, n2, n3))
                except (ConnectionError, TimeoutError, ProtocolException) as e:
                    log.warning('Error reading {}: {}'.format(comment, e))
                    if not ctx.is_open():
                        raise
            # 08_11
            for x in self.fn_08_11_:
                name = x['tag']
                factor = x['f']
                bwri = calc_bwri(x['bwri'])

                comment = 'Read {}'.format(name)
                try:
                    self.read_fn_08_11(ctx, bwri, factor, comment, name)
                except (ConnectionError, TimeoutError, ProtocolException) as ex:
                    log.warning('Error reading {}: {}'.format(comment, ex))
                    if not ctx.is_open():
                        raise

            comment = 'Read freq'
            try:
                self.read_freq(ctx, comment)
            except (ConnectionError, TimeoutError, ProtocolException) as ex:
                log.warning('Error reading {}: {}'.format(comment, ex))
                if not ctx.is_open():
                    raise

        except (ConnectionError, TimeoutError, ProtocolException) as e:
            log.warning(e)

        finally:
            for name, con_val in self.values_.items():
                if not con_val.is_read():
                    con_val.invalidate(Quality.QUALITY_COMM)
            self.timeout_general_.start()

    def open_channel(self, ctx):
        # access level
        req = bytes([self.addr_, 0x01, 0x01]) + self.password_

        ctx.send_rcv(req, None, 'Open channel', self.log())

    @staticmethod
    # None => энергия маскирована (не поддерживается)
    # rval: вт*ч
    def parse_energy_4_bytes(d: (bytes, bytearray), off: int) -> (int, None):
        assert isinstance(d, (bytes, bytearray)), d
        assert isinstance(off, int), off
        assert off + 4 <= len(d), (d, off)

        b0 = d[off + 0]
        b1 = d[off + 1]
        b2 = d[off + 2]
        b3 = d[off + 3]

        if b0 == 0xff and b1 == 0xff and b2 == 0xff and b3 == 0xff:
            return None

        return b2 | (b3 << 8) | (b0 << 16) | (b1 << 24)

    def read_energy(self, ctx, array_nb, comment, tpl):
        log = self.log()

        month_nb = 0
        tariff_nb = 0  # 0 - сумма тарифов

        req = bytes([self.addr_, 0x05, (array_nb << 4) | month_nb, tariff_nb])

        ctx.send_rcv(req, 16, comment, log)

        eap_wth = MerkDevice.parse_energy_4_bytes(ctx.data(), 0)
        ean_wth = MerkDevice.parse_energy_4_bytes(ctx.data(), 4)
        erp_wth = MerkDevice.parse_energy_4_bytes(ctx.data(), 8)
        ern_wth = MerkDevice.parse_energy_4_bytes(ctx.data(), 12)

        eap_kwth = 0.0 if eap_wth is None else eap_wth / 1000.0
        ean_kwth = 0.0 if ean_wth is None else ean_wth / 1000.0
        erp_kwth = 0.0 if erp_wth is None else erp_wth / 1000.0
        ern_kwth = 0.0 if ern_wth is None else ern_wth / 1000.0

        log.debug('Eap={} Ean={} Erp={} Ern={} [kwth]'.format(eap_kwth, ean_kwth, erp_kwth, ern_kwth))

        eap_name, ean_name, erp_name, ern_name = tpl

        self.values_[eap_name].set(eap_kwth)
        self.values_[ean_name].set(ean_kwth)
        self.values_[erp_name].set(erp_kwth)
        self.values_[ern_name].set(ern_kwth)

    # bwri - код чтения
    # factor - коэффициент (в Меркурии все измерения масштабируются статическим коэффициентом)
    # tpl - имена переменных, куда записать значения
    # pq_sign - 'P', 'Q', None
    def read_pqs(self, ctx, bwri, factor, comment, tpl, pq_sign):

        assert 0 <= bwri <= 0xff, bwri
        assert isinstance(tpl, tuple) and len(tpl) == 4, tpl
        assert pq_sign == 'P' or pq_sign == 'Q' or pq_sign is None, pq_sign

        def extract_pqs(d, off) -> int:
            b2 = d[off + 0]
            b1 = d[off + 1]
            b4 = d[off + 2]
            b3 = d[off + 3]

            b1_data = b1 & 0x3F     # маскируем старшие 2 разряда
            
            # старший разряд - направление P (0 = прямое, 1 = обратное)
            p_sign = 1 if (b1 & 0x80) == 0 else -1
            
            # 2й слева разряд - направление Q (0 = прямое, 1 = обратное)
            q_sign = 1 if (b1 & 0x40) == 0 else -1

            val = b4 | (b3 << 8) | (b2 << 16) | (b1_data << 24)
            
            if pq_sign == 'P':
                return val * p_sign

            if pq_sign == 'Q':
                return val * q_sign
            
            return val


        log = self.log()

        # 2.3.14.1
        req = bytes([self.addr_, 0x08, 0x14, bwri])
        ctx.send_rcv(req, 16, comment, log)

        sum_raw = extract_pqs(ctx.data(), 0)
        ph1_raw = extract_pqs(ctx.data(), 4)
        ph2_raw = extract_pqs(ctx.data(), 8)
        ph3_raw = extract_pqs(ctx.data(), 12)

        sum_phys = sum_raw * factor
        ph1_phys = ph1_raw * factor
        ph2_phys = ph2_raw * factor
        ph3_phys = ph3_raw * factor

        log.debug('%s: sum_raw=%s, ph1_raw=%s, ph2_raw=%s, ph3_raw=%s, sum=%s, ph1=%s, ph2=%s, ph3=%s',
                  comment, sum_raw, ph1_raw, ph2_raw, ph3_raw, sum_phys, ph1_phys, ph2_phys, ph3_phys)

        sum_name, ph1_name, ph2_name, ph3_name = tpl

        self.values_[sum_name].set(sum_phys)
        self.values_[ph1_name].set(ph1_phys)
        self.values_[ph2_name].set(ph2_phys)
        self.values_[ph3_name].set(ph3_phys)
        
    def read_uicos(self, ctx, bwri, factor, comment, tpl):

        assert 0 <= bwri <= 0xff, bwri
        assert isinstance(tpl, tuple) and len(tpl) == 3, tpl

        def extract_uicos(d, off) -> int:
            b1 = d[off + 0]
            b3 = d[off + 1]
            b2 = d[off + 2]

            return (b1 << 16) | (b2 << 8) | b3


        log = self.log()

        # 2.3.14.2
        req = bytes([self.addr_, 0x08, 0x14, bwri])
        ctx.send_rcv(req, 9, comment, log)

        ph1_raw = extract_uicos(ctx.data(), 0)
        ph2_raw = extract_uicos(ctx.data(), 3)
        ph3_raw = extract_uicos(ctx.data(), 6)

        ph1_phys = ph1_raw * factor
        ph2_phys = ph2_raw * factor
        ph3_phys = ph3_raw * factor

        log.debug('%s: ph1_raw=%s, ph2_raw=%s, ph3_raw=%s, ph1=%s, ph2=%s, ph3=%s', comment, ph1_raw, ph2_raw, ph3_raw, ph1_phys, ph2_phys, ph3_phys)

        ph1_name, ph2_name, ph3_name = tpl

        self.values_[ph1_name].set(ph1_phys)
        self.values_[ph2_name].set(ph2_phys)
        self.values_[ph3_name].set(ph3_phys)
        
    def read_fn_08_11(self, ctx, bwri, factor, comment, name):
        assert 0 <= bwri <= 0xff, bwri
        assert isinstance(name, str), name

        def __extract(d) -> int:
            b1 = d[0]
            b3 = d[1]
            b2 = d[2]

            return (b1 << 16) | (b2 << 8) | b3


        log = self.log()

        # 2.3.14.2
        req = bytes([self.addr_, 0x08, 0x11, bwri])
        ctx.send_rcv(req, 3, comment, log)

        val_raw = __extract(ctx.data())

        val_phys = val_raw * factor

        log.debug('%s: raw=%s, ph=%s', comment, val_raw, val_phys)

        self.values_[name].set(val_phys)
        
    def read_freq(self, ctx, comment):

        log = self.log()

        # 2.3.14.4
        req = bytes([self.addr_, 0x08, 0x14, 0x40])
        ctx.send_rcv(req, 3, comment, log)

        d = ctx.data()
        b1 = d[0]
        b3 = d[1]
        b2 = d[2]

        freq_raw = b3 | (b2 << 8) | (b1 << 16)
        freq_phys = freq_raw * 0.01

        log.debug('%s: raw=%s, phys=%s', comment, freq_raw, freq_phys)

        self.values_[FREQ_TAG_NAME].set(freq_phys)
        
    def read_status(self, ctx, comment):

        log = self.log()

        # 2.3.10
        req = bytes([self.addr_, 0x08, 0x0a])
        ctx.send_rcv(req, 6, comment, log)

        d = ctx.data()
        
        big_val = (d[0] << 40) | (d[1] << 32) | (d[2] << 24) | (d[3] << 16) | (d[4] << 8) | d[5]
        
        log.debug('%s: status=%s', comment, big_val)

        self.values_[STATUS_TAG_NAME].set(big_val)




# Raises ValueError
def bcd_to_int(bcd: int) -> int:
    assert isinstance(bcd, int), bcd
    
    lo = bcd & 0x0f
    hi = (bcd >> 4) & 0x0f

    if lo > 9 or hi > 9:
        raise ValueError('Value 0x{:x} is not a valid BCD'.format(bcd))

    return 10 * hi + lo

def make_word(high_byte: int, low_byte: int) -> int:
    assert isinstance(high_byte, int) and (0 <= high_byte <= 0xff), high_byte
    assert isinstance(low_byte, int) and 0 <= low_byte <= 0xff, low_byte
    
    return (high_byte << 8) | low_byte
    

def hi_byte(val: int) -> int:
    assert isinstance(val, int), val
    return (val >> 8) & 0xff


def lo_byte(val: int) -> int:
    assert isinstance(val, int), val
    return val & 0xff
    


import struct

from ..lib.stddef import Quality, ValueType
from ..lib import stdsvcdef
from ..lib.xml_reader import XmlReader, LoadException
from ..lib import modbus

from . import time_tools


class MbFrame:
    def __init__(self, my_xml_node, proposed_aux_name, parent_dev, parent_log):
        p = XmlReader(my_xml_node)

        self.parent_dev_ = parent_dev
        self.aux_name_ = p.get_str('name', proposed_aux_name)
        self.log_ = parent_log.getChild(self.aux_name_)

    def aux_name(self):
        return self.aux_name_

    def log(self):
        return self.log_

    # def set_tag(self, tag_name_or_id: (int, str), value, quality: int) -> None:
    #    assert self.pipe_set_
    #    self.pipe_set_.set_tag(tag_name_or_id, value, quality)

    def device(self):
        return self.parent_dev_

    def empty(self):
        raise NotImplementedError()

    def read_write(self, ctx):
        raise NotImplementedError()

    def sync(self, services: dict):
        raise NotImplementedError()

    def on_channel_closed(self):
        raise NotImplementedError()


class MbFrameReg(MbFrame):
    def __init__(self, xml_node, proposed_aux_name, parent_dev, parent_log):

        super().__init__(xml_node, proposed_aux_name, parent_dev, parent_log)

        p = XmlReader(xml_node)

        self.fn_nb_ = p.get_int('fn', default=3, min_val=3, max_val=4)
        self.addr_ = None  # p.get_int('fn', default=-1, min=0, max=0xffff)
        self.nb_words_ = None

        read_period = p.get_time_secs('readPeriod', default=10.0, min_val=0.0, max_val=60.0)
        self.read_timeout_ = time_tools.Timeout(read_period, expired=True)

        self.ties_ = []
        self.load_ties_(xml_node)

        if not self.ties_:
            return

        # calc addr & size
        last_addr = None
        for t in self.ties_:
            if self.addr_ is None or self.addr_ > t.addr():
                self.addr_ = t.addr()

            tie_last_addr = t.addr() + t.size_words() - 1
            if last_addr is None or last_addr < tie_last_addr:
                last_addr = tie_last_addr

        assert last_addr > self.addr_

        self.nb_words_ = last_addr - self.addr_ + 1
        if self.nb_words_ > modbus.MAX_READ_WORDS_NB:
            raise LoadException('Frame size {} exceeds maximum allowed {}'.format(self.nb_words_, modbus.MAX_READ_WORDS_NB), xml_node)

    def empty(self):
        return len(self.ties_) == 0

    def sync(self, services: dict):
        assert isinstance(services, dict), services
        for t in self.ties_:
            t.sync(services)

    def load_ties_(self, xml_node):
        for node in xml_node:
            if node.tag != 'Link':
                raise LoadException('Unknown node name: {}'.format(node.tag), node)

            p = XmlReader(node)

            # tag_full_name = self.device().get_tag_full_name(p.get_str('name'))

            # TagInfo ti;
            # tie: MbFrameRegTie
            fmt_str = p.get_str('fmt')
            fmt = modbus.MB_FORMATS_DICT.get(fmt_str, None)
            if fmt:
                tie = MbFrameRegTieNum(p, fmt, self.device().branch(), self.log())
            else:
                if fmt_str == 'bool_sp':
                    pass  # tie = new MbWordFrameTie_Bool_SP( node, ti, &m_Address, log, &ok );
                elif fmt_str == 'bool_dp':
                    pass  # tie = new MbWordFrameTie_Bool_DP( node, ti, &m_Address, log, &ok );
                elif fmt_str == 'bool_cmp':
                    pass  # tie = new MbWordFrameTie_Bool_RegValueCompare( node, ti, &m_Address, log, &ok );
                else:
                    raise LoadException('Unknown format: "{}"'.format(fmt_str), node)

            assert tie
            # if ( tie->isWritable() )
            #    getDevice()->SubscribeTag( ti.id, this, ties_.size(), 0 );
            self.ties_.append(tie)

    def read_write(self, ctx):
        if not self.read_timeout_.is_expired():
            return

        log = self.log()

        log.debug('Read frame')

        try:
            data_bytes = ctx.read_words(self.device().slave_nb(), self.fn_nb_, self.addr_, self.nb_words_, self.aux_name(), log)
            assert isinstance(data_bytes, (bytes, bytearray)), data_bytes
            assert len(data_bytes) == self.nb_words_ * 2, data_bytes

            for tie in self.ties_:
                tie.on_read_complete(data_bytes, 2 * (tie.addr() - self.addr_))

        except (ConnectionError, TimeoutError, modbus.MbException, modbus.MbError) as e:
            log.warning(e)
            for tie in self.ties_:
                tie.invalidate(Quality.QUALITY_COMM)
        finally:
            self.read_timeout_.start()

    def on_channel_closed(self):
        for tie in self.ties_:
            tie.invalidate(Quality.QUALITY_COMM)


class MbFrameRegTie:
    def sync(self, services: dict):
        raise NotImplementedError()

    def on_read_complete(self, data_bytes: bytearray, my_offset: int):
        raise NotImplementedError()

    def invalidate(self, quality: int):
        raise NotImplementedError()


class MbFrameRegTieNum(MbFrameRegTie, stdsvcdef.IFindTagHandler):
    # mb_fmt: see vctools.modbus_formats
    def __init__(self, p: XmlReader, mb_fmt: tuple, branch: str, log):
        assert isinstance(p, XmlReader), p
        assert isinstance(mb_fmt, tuple), mb_fmt
        assert isinstance(branch, str), branch

        fmt_id, fmt_size_words, fmt_min_n, fmt_max_n, fmt_struct_format, *r = mb_fmt    # pylint: disable=unused-variable

        self.fmt_size_words_ = fmt_size_words
        self.fmt_map_string_ = fmt_struct_format
        self.tag_short_name_ = p.get_str('name')
        self.tag_full_name_ = branch + '.' + self.tag_short_name_
        self.abs_addr_ = p.get_int('addr', min_val=0, max_val=(0xffff - fmt_size_words + 1))

        self.log_ = log
        # self.pipe_set_: services_defs.ISetValueService = None
        self.pipe_set_ = None

        self.value_ = None
        self.quality_ = Quality.QUALITY_NOT_INIT

    def addr(self):
        return self.abs_addr_

    def size_words(self):
        return self.fmt_size_words_

    def sync(self, services: dict):
        assert isinstance(services, dict), services

        assert 'set' in services, services
        assert isinstance(services['set'], stdsvcdef.ISetValueService)
        self.pipe_set_ = services['set']

        assert 'find' in services, services
        assert isinstance(services['find'], stdsvcdef.IFindTagService)
        services['find'].find_tag(self.tag_full_name_, self)

    def on_find_tag_ok(self, tag_id: int, tag_name: str, vt: str, user_data):
        if ValueType(vt) in (ValueType.VT_INT, ValueType.VT_FLOAT):
            # pipe.subscribe(tag_name, self)
            self.pipe_set_.set_tag(tag_name, self.value_, self.quality_)
        else:
            raise Warning('Cant link tag {} [{}]. Only support int,float'.format(tag_name, vt))

    def on_find_tag_err(self, tag_name_or_id: (int, str), user_data):
        raise Warning('Cant link tag {}: tag not found'.format(self.tag_full_name_))

    def on_read_complete(self, data_bytes: bytearray, my_offset: int):
        value, = struct.unpack_from(self.fmt_map_string_, data_bytes, my_offset)

        if self.value_ != value or self.quality_ != 0:
            self.value_ = value
            self.quality_ = 0
            self.log_.debug('Tie updated {}={}[{}]'.format(self.tag_full_name_, self.value_, self.quality_))
            self.pipe_set_.set_tag(self.tag_full_name_, self.value_)

    def invalidate(self, quality: int):
        if self.value_ is not None or self.quality_ != quality:
            self.value_ = None
            self.quality_ = quality
            self.log_.debug('Tie invalidated {}={}[{}]'.format(self.tag_full_name_, self.value_, self.quality_))
            self.pipe_set_.set_tag(self.tag_full_name_, None, self.quality_)



#!/usr/bin/python3

import time

from ..lib.bin import bytes_to_str

from . import time_tools


class BaseCtx:
    def __init__(self, channel):
        self.silence_sec = 0.0
        self.timeout_sec = 1.0
        #self.abort_cb = None

        self.log_ = None
        self.channel_ = channel
        self.comment_ = None

    def log(self):
        return self.log_

    def close_channel(self):
        self.channel_.safe_close()

    def is_open(self):
        return self.channel_.is_open()

    def test_abort(self):
        #if self.abort_cb and self.abort_cb():
        #    raise SystemExit()
        pass

    # None  => continue listen
    # Any   => return as result of send_rcv()
    # Exception => if channel is open, retry (swallow exception), else fail (re-raise)
    def data_cb(self, data: (bytes, bytearray)):
        raise NotImplementedError()

    # rval:
    #
    #   success         =>  rval from data_cb()
    #
    #   * SystemExit        =>  aborted by abort_cb()
    #   * ConnectionError   =>  Any problems with channel
    #   * TimeoutError
    #   * AnyException from data_cb()
    def _send_rcv_once(self, data_to_send: (bytes, bytearray)):
        assert isinstance(data_to_send, (bytes, bytearray)), data_to_send

        if self.silence_sec > 0.0:
            self.test_abort()  # *SystemExit
            time.sleep(self.silence_sec)
            self.test_abort()  # *SystemExit

        ch = self.channel_

        ch.reset_rcv_buf_x(self.log_)  # *ConnectionError

        if self.log_:
            if self.comment_:
                self.log_.debug('S> {} [{}]'.format(bytes_to_str(data_to_send), self.comment_))
            else:
                self.log_.debug('S> {}'.format(bytes_to_str(data_to_send)))

        ch.write_all_x(data_to_send, self.log_)  # *ConnectionError

        timeout = time_tools.Timeout(self.timeout_sec)

        rcvd = b''

        while not timeout.is_expired():
            self.test_abort()  # *SystemExit

            # b'' if not data
            portion = ch.read_x(timeout.remain_sec(), self.log_)  # *ConnectionError
            assert isinstance(portion, (bytes, bytearray)), portion

            if portion:
                if self.log_:
                    self.log_.debug('R< {}'.format(bytes_to_str(portion)))
                rcvd += portion

                r_process = self.data_cb(rcvd)

                if r_process is not None:
                    return r_process

        if self.comment_:
            raise TimeoutError('{}: Timeout'.format(self.comment_))
        else:
            raise TimeoutError('Timeout')

    # rval:
    #
    #   success         =>  rval from data_cb()
    #
    #   * SystemExit        =>  aborted by abort_cb()
    #   * ConnectionError   =>  Any problems with channel
    #   * TimeoutError
    #   * AnyException from data_cb()
    def send_rcv_raw(self, data_to_send: (bytes, bytearray), comment_n: str, log, nb_attempts: int = 2):
        assert isinstance(data_to_send, (bytes, bytearray)), data_to_send
        assert comment_n is None or isinstance(comment_n, str), comment_n
        assert isinstance(nb_attempts, int), nb_attempts

        try:
            self.log_ = log
            self.comment_ = comment_n

            cr_attempt_nb = 1
            while True:
                try:
                    return self._send_rcv_once(data_to_send)
                except Exception as e:
                    if not self.channel_.is_open():
                        raise e

                    cr_attempt_nb += 1
                    if cr_attempt_nb > nb_attempts:
                        raise e

        finally:
            self.log_ = None
            self.comment_ = None
